#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe


from typing import Any

import torch
from ax.adapter.registry import Generators
from ax.api.utils.generation_strategy_dispatch import choose_generation_strategy
from ax.api.utils.structs import GenerationStrategyDispatchStruct
from ax.core.trial import Trial
from ax.core.trial_status import TrialStatus
from ax.generation_strategy.center_generation_node import CenterGenerationNode
from ax.generation_strategy.transition_criterion import MinTrials
from ax.generators.torch.botorch_modular.surrogate import ModelConfig, SurrogateSpec
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_experiment,
    get_experiment_with_observations,
)
from ax.utils.testing.mock import mock_botorch_optimize
from ax.utils.testing.utils import run_trials_with_gs
from botorch.models.fully_bayesian import SaasFullyBayesianSingleTaskGP
from pyre_extensions import assert_is_instance, none_throws


class TestDispatchUtils(TestCase):
    def test_choose_gs_random_search(self) -> None:
        struct_kws_cases: dict[str, dict[str, Any]] = {
            "use_center_false": {"initialize_with_center": False},
            "use_center_true": {"initialize_with_center": True},
            "default": {},
        }
        use_center_cases = {
            "use_center_false": False,
            "use_center_true": True,
            "default": True,
        }
        for case, struct_kws in struct_kws_cases.items():
            with self.subTest(case=case):
                struct = GenerationStrategyDispatchStruct(
                    method="random_search", **struct_kws
                )
                use_center = use_center_cases[case]
                gs = choose_generation_strategy(struct=struct)
                self.assertEqual(len(gs._nodes), 1 + use_center)
                if use_center:
                    self.assertIsInstance(gs._nodes[0], CenterGenerationNode)
                    self.assertEqual(gs.name, "Center+QuasiRandomSearch")
                else:
                    self.assertEqual(gs.name, "QuasiRandomSearch")
                sobol_node = gs._nodes[-1]
                self.assertEqual(len(sobol_node.generator_specs), 1)
                sobol_spec = sobol_node.generator_specs[0]
                self.assertEqual(sobol_spec.generator_enum, Generators.SOBOL)
                self.assertEqual(sobol_spec.model_kwargs, {"seed": None})
                self.assertEqual(sobol_node._transition_criteria, [])
                # Make sure it generates.
                run_trials_with_gs(
                    experiment=get_branin_experiment(), gs=gs, num_trials=3
                )

    @mock_botorch_optimize
    def test_choose_gs_fast_with_options(self) -> None:
        struct = GenerationStrategyDispatchStruct(
            method="fast",
            initialization_budget=3,
            initialization_random_seed=0,
            use_existing_trials_for_initialization=False,
            min_observed_initialization_trials=4,
            allow_exceeding_initialization_budget=True,
            torch_device="cpu",
            initialize_with_center=True,
        )
        gs = choose_generation_strategy(struct=struct)
        self.assertEqual(len(gs._nodes), 3)
        self.assertEqual(gs.name, "Center+Sobol+MBM:fast")
        # Check the center node.
        center_node = assert_is_instance(gs._nodes[0], CenterGenerationNode)
        self.assertEqual(center_node.next_node_name, "Sobol")
        # Check the Sobol node & TC.
        sobol_node = gs._nodes[1]
        self.assertTrue(sobol_node.should_deduplicate)
        self.assertEqual(len(sobol_node.generator_specs), 1)
        sobol_spec = sobol_node.generator_specs[0]
        self.assertEqual(sobol_spec.generator_enum, Generators.SOBOL)
        self.assertEqual(sobol_spec.model_kwargs, {"seed": 0})
        expected_tc = [
            MinTrials(
                threshold=2,
                transition_to="MBM",
                block_gen_if_met=False,
                block_transition_if_unmet=True,
                use_all_trials_in_exp=False,
            ),
            MinTrials(
                threshold=4,
                transition_to="MBM",
                block_gen_if_met=False,
                block_transition_if_unmet=True,
                use_all_trials_in_exp=True,
                only_in_statuses=[TrialStatus.COMPLETED],
                count_only_trials_with_data=True,
            ),
        ]
        self.assertEqual(sobol_node._transition_criteria, expected_tc)
        # Check the MBM node.
        mbm_node = gs._nodes[2]
        self.assertTrue(mbm_node.should_deduplicate)
        self.assertEqual(len(mbm_node.generator_specs), 1)
        mbm_spec = mbm_node.generator_specs[0]
        self.assertEqual(mbm_spec.generator_enum, Generators.BOTORCH_MODULAR)
        expected_ss = SurrogateSpec(model_configs=[ModelConfig(name="MBM defaults")])
        self.assertEqual(
            mbm_spec.model_kwargs,
            {"surrogate_spec": expected_ss, "torch_device": torch.device("cpu")},
        )
        self.assertEqual(mbm_node._transition_criteria, [])
        # Experiment with 2 observations. We should still generate 4 Sobol trials.
        experiment = get_experiment_with_observations([[1.0], [2.0]])
        # Mark the existing trials as manual to prevent them from counting for Sobol.
        for trial in experiment.trials.values():
            none_throws(
                assert_is_instance(trial, Trial).generator_run
            )._model_key = "Manual"
        # Generate 5 trials and make sure they're from the correct nodes.
        run_trials_with_gs(experiment=experiment, gs=gs, num_trials=5)
        self.assertEqual(len(experiment.trials), 7)
        for trial in experiment.trials.values():
            model_key = none_throws(
                assert_is_instance(trial, Trial).generator_run
            )._model_key
            if trial.index < 2:
                self.assertEqual(model_key, "Manual")
            elif trial.index == 2:
                self.assertEqual(model_key, "CenterOfSearchSpace")
            elif trial.index < 5:
                self.assertEqual(model_key, "Sobol")
            else:
                self.assertEqual(model_key, "BoTorch")

    def test_choose_gs_quality_with_options(self) -> None:
        struct = GenerationStrategyDispatchStruct(
            method="quality",
            initialization_budget=3,
            initialization_random_seed=0,
            use_existing_trials_for_initialization=False,
            min_observed_initialization_trials=4,
            allow_exceeding_initialization_budget=True,
            torch_device="cpu",
            initialize_with_center=True,
        )
        gs = choose_generation_strategy(struct=struct)
        self.assertEqual(len(gs._nodes), 3)
        self.assertEqual(gs.name, "Center+Sobol+MBM:quality")

        # Check the MBM node.
        mbm_node = gs._nodes[2]
        self.assertTrue(mbm_node.should_deduplicate)
        self.assertEqual(len(mbm_node.generator_specs), 1)
        mbm_spec = mbm_node.generator_specs[0]
        self.assertEqual(mbm_spec.generator_enum, Generators.BOTORCH_MODULAR)
        expected_ss = SurrogateSpec(
            model_configs=[
                ModelConfig(
                    botorch_model_class=SaasFullyBayesianSingleTaskGP,
                    model_options={"use_input_warping": True},
                    mll_options={
                        "disable_progbar": True,
                    },
                    name="WarpedSAAS",
                )
            ]
        )
        self.assertEqual(
            mbm_spec.model_kwargs,
            {"surrogate_spec": expected_ss, "torch_device": torch.device("cpu")},
        )
        self.assertEqual(mbm_node._transition_criteria, [])

    def test_choose_gs_no_initialization(self) -> None:
        struct = GenerationStrategyDispatchStruct(
            method="fast", initialization_budget=0
        )
        gs = choose_generation_strategy(struct=struct)
        self.assertEqual(len(gs._nodes), 1)
        self.assertEqual(gs.name, "MBM:fast")
        mbm_node = gs._nodes[0]
        self.assertEqual(mbm_node.node_name, "MBM")

    def test_choose_gs_center_only_initialization(self) -> None:
        struct = GenerationStrategyDispatchStruct(
            method="fast", initialization_budget=1, initialize_with_center=True
        )
        gs = choose_generation_strategy(struct=struct)
        self.assertEqual(len(gs._nodes), 2)
        self.assertEqual(gs.name, "Center+MBM:fast")
        center_node = gs._nodes[0]
        self.assertEqual(center_node.node_name, "CenterOfSearchSpace")
        mbm_node = gs._nodes[1]
        self.assertEqual(mbm_node.node_name, "MBM")

    def test_choose_gs_single_sobol_initialization(self) -> None:
        struct = GenerationStrategyDispatchStruct(
            method="fast", initialization_budget=1, initialize_with_center=False
        )
        gs = choose_generation_strategy(struct=struct)
        self.assertEqual(len(gs._nodes), 2)
        self.assertEqual(gs.name, "Sobol+MBM:fast")
        sobol_node = gs._nodes[0]
        self.assertEqual(sobol_node.node_name, "Sobol")
        mbm_node = gs._nodes[1]
        self.assertEqual(mbm_node.node_name, "MBM")



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from dataclasses import replace

import torch
from ax.benchmark.benchmark_test_functions.botorch_test import BoTorchTestFunction
from ax.utils.common.testutils import TestCase
from botorch.test_functions.multi_objective import BraninCurrin
from botorch.test_functions.synthetic import ConstrainedHartmann, Hartmann


class TestBoTorchTestFunction(TestCase):
    def setUp(self) -> None:
        super().setUp()
        botorch_base_test_functions = {
            "base Hartmann": Hartmann(dim=6),
            "negated Hartmann": Hartmann(dim=6, negate=True),
            "constrained Hartmann": ConstrainedHartmann(dim=6),
            "negated constrained Hartmann": ConstrainedHartmann(dim=6, negate=True),
        }
        self.outcome_names = {
            "Hartmann": ["y"],
            "constrained Hartmann": ["y", "c"],
        }
        self.botorch_test_functions = {
            k: BoTorchTestFunction(
                botorch_problem=v,
                outcome_names=self.outcome_names[
                    "constrained Hartmann" if "constrained" in k else "Hartmann"
                ],
            )
            for k, v in botorch_base_test_functions.items()
        }

    def test_negation(self) -> None:
        params = {f"x{i}": 0.5 for i in range(6)}
        evaluate_true_results = {
            k: v.evaluate_true(params) for k, v in self.botorch_test_functions.items()
        }
        self.assertEqual(
            evaluate_true_results["base Hartmann"],
            evaluate_true_results["constrained Hartmann"][0],
        )
        self.assertEqual(
            evaluate_true_results["base Hartmann"],
            -evaluate_true_results["negated Hartmann"],
        )
        self.assertEqual(
            evaluate_true_results["negated Hartmann"],
            evaluate_true_results["negated constrained Hartmann"][0],
        )
        self.assertEqual(
            evaluate_true_results["constrained Hartmann"][1],
            evaluate_true_results["negated constrained Hartmann"][1],
        )

    def test_raises_for_botorch_attrs(self) -> None:
        msg = "noise should be set on the `BenchmarkRunner`, not the test function."
        with self.assertRaisesRegex(ValueError, msg):
            BoTorchTestFunction(
                botorch_problem=Hartmann(dim=6, noise_std=0.1),
                outcome_names=self.outcome_names["Hartmann"],
            )
        with self.assertRaisesRegex(ValueError, msg):
            BoTorchTestFunction(
                botorch_problem=ConstrainedHartmann(dim=6, constraint_noise_std=0.1),
                outcome_names=self.outcome_names["constrained Hartmann"],
            )

    def test_tensor_shapes(self) -> None:
        params = {f"x{i}": 0.5 for i in range(6)}
        evaluate_true_results = {
            k: v.evaluate_true(params) for k, v in self.botorch_test_functions.items()
        }
        evaluate_true_results["BraninCurrin"] = BoTorchTestFunction(
            botorch_problem=BraninCurrin(), outcome_names=["f1", "f2"]
        ).evaluate_true({"x0": 0.0, "x1": 0.0})
        expected_len = {
            "base Hartmann": 1,
            "constrained Hartmann": 2,
            "negated Hartmann": 1,
            "negated constrained Hartmann": 2,
            "BraninCurrin": 2,
        }
        for name, result in evaluate_true_results.items():
            with self.subTest(name=name):
                self.assertEqual(result.dtype, torch.double)
                self.assertEqual(result.shape, torch.Size([expected_len[name], 1]))

    def test_input_dimensions(self) -> None:
        test_function = self.botorch_test_functions["base Hartmann"]
        with self.assertRaisesRegex(ValueError, "Expected 6 parameters, got 7"):
            test_function.evaluate_true({f"x{i}": 0.5 for i in range(7)})

        with self.assertRaisesRegex(ValueError, "Expected 7 parameters, got 6"):
            test_function = replace(test_function, dummy_param_names={"ignore me"})
            test_function.evaluate_true({f"x{i}": 0.5 for i in range(6)})

    def test_n_steps_is_one(self) -> None:
        self.assertEqual(self.botorch_test_functions["base Hartmann"].n_steps, 1)

    def test_dummy_dimensions(self) -> None:
        test_function = self.botorch_test_functions["base Hartmann"]
        embedded_test_function = replace(test_function, dummy_param_names={"ignore me"})
        params = {f"x{i}": 0.5 for i in range(6)}
        embedded_params = {**params, "ignore me": 0.5}
        self.assertEqual(
            test_function.evaluate_true(params=params),
            embedded_test_function.evaluate_true(params=embedded_params),
        )

    def test_with_steps(self) -> None:
        ha_params = {f"x{i}": 0.5 for i in range(6)}
        br_params = {"x0": 0.0, "x1": 0.0}
        for name, botorch_problem, outcome_names, params in [
            ("Unconstrained", Hartmann(dim=6), ["y"], ha_params),
            ("Constrained", ConstrainedHartmann(dim=6), ["y", "c"], ha_params),
            ("Moo", BraninCurrin(), ["y1", "y2"], br_params),
        ]:
            for n_steps in [1, 3]:
                with self.subTest(name=name):
                    test_function = BoTorchTestFunction(
                        outcome_names=outcome_names,
                        botorch_problem=botorch_problem,
                        n_steps=n_steps,
                    )
                    self.assertEqual(test_function.n_steps, n_steps)
                    result = test_function.evaluate_true(params=params)
                    self.assertEqual(
                        result.shape, torch.Size([len(outcome_names), n_steps])
                    )
                    if n_steps == 2:
                        # data is simply repeated down the step dimension
                        self.assertEqual(result[0, 0].item(), result[0, 1].item())



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from collections.abc import Callable, Generator, Iterable, Mapping
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime
from enum import IntEnum
from logging import LoggerAdapter
from time import sleep
from typing import Any, cast, NamedTuple

import ax.service.utils.early_stopping as early_stopping_utils
from ax.adapter.adapter_utils import get_fixed_features_from_experiment
from ax.adapter.base import Adapter
from ax.core.base_trial import BaseTrial, TrialStatus
from ax.core.experiment import Experiment
from ax.core.generator_run import GeneratorRun
from ax.core.metric import Metric, MetricFetchE, MetricFetchResult
from ax.core.multi_type_experiment import (
    filter_trials_by_type,
    get_trial_indices_for_statuses,
    MultiTypeExperiment,
)
from ax.core.runner import Runner
from ax.core.utils import get_pending_observation_features_based_on_trial_status
from ax.exceptions.core import (
    AxError,
    DataRequiredError,
    OptimizationComplete,
    UnsupportedError,
    UserInputError,
)
from ax.exceptions.generation_strategy import (
    AxGenerationException,
    MaxParallelismReachedException,
    OptimizationConfigRequired,
)
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.service.utils.analysis_base import AnalysisBase
from ax.service.utils.best_point_mixin import BestPointMixin
from ax.service.utils.orchestrator_options import OrchestratorOptions, TrialType
from ax.service.utils.with_db_settings_base import DBSettings, WithDBSettingsBase
from ax.utils.common.constants import Keys
from ax.utils.common.executils import retry_on_exception
from ax.utils.common.logger import (
    build_file_handler,
    get_logger,
    make_indices_str,
    set_ax_logger_levels,
)
from ax.utils.common.timeutils import current_timestamp_in_millis
from pyre_extensions import assert_is_instance, none_throws


NOT_IMPLEMENTED_IN_BASE_CLASS_MSG = """ \
This method is not implemented in the base `Orchestrator` class. \
If this functionality is desired, specify the method in the \
Orchestrator subclass.
"""
GS_TYPE_MSG = "This optimization run uses a '{gs_name}' generation strategy."
OPTIMIZATION_COMPLETION_MSG = """Optimization completed with total of {num_trials} \
trials attached to the underlying Ax experiment '{experiment_name}'.
"""
FAILURE_EXCEEDED_MSG = (
    "Failure rate exceeds the tolerated trial failure rate of {f_rate} (at least "
    "{n_failed} out of first {n_ran} trials failed or were abandoned). Checks are "
    "triggered both at the end of an optimization and if at least {min_failed} trials "
    "have either failed, or have been abandoned, potentially automatically due to "
    "issues with the trial."
)


# Wait time b/w reports will not exceed 15 mins.
MAX_SECONDS_BETWEEN_REPORTS = 900


class OptimizationResult(NamedTuple):  # TODO[T61776778]
    pass  # TBD


class OrchestratorInternalError(AxError):
    """Error that indicates an error within the `Orchestrator` logic."""

    pass


class FailureRateExceededError(AxError):
    """Error that indicates the optimization was aborted due to excessive
    failure rate.
    """

    pass


NO_RETRY_EXCEPTIONS: tuple[type[Exception], ...] = (
    cast(type[Exception], OrchestratorInternalError),
    cast(type[Exception], NotImplementedError),
    cast(type[Exception], UnsupportedError),
)


class OutputPriority(IntEnum):
    """Priority of a message. Messages with higher priority will be shown first, and
    messages with the same priority will be sorted alphabetically."""

    NOTSET = 0
    DEBUG = 10
    INFO = 20
    TOPLINE = 30
    WARNING = 40
    ERROR = 50


@dataclass
class MessageOutput:
    """Message to be shown in the output of the orchestrator."""

    text: str
    priority: OutputPriority | int

    def __str__(self) -> str:
        return self.text

    def __repr__(self) -> str:
        return f"MessageOutput(text={self.text}, priority={self.priority})"

    def append(self, text: str) -> None:
        """Append text to the text of an existing message."""
        self.text += text


class Orchestrator(AnalysisBase, BestPointMixin):
    """Closed-loop manager class for Ax optimization.

    Attributes:
        experiment: Experiment, in which results of the optimization
            will be recorded.
        generation_strategy: Generation strategy for the optimization,
            describes models that will be used in optimization.
        options: `OrchestratorOptions` for this Orchestrator instance.
        db_settings: Settings for saving and reloading the underlying experiment
            to a database. Expected to be of type
            ax.storage.sqa_store.structs.DBSettings and require SQLAlchemy.
        _skip_experiment_save: If True, Orchestrator will not re-save the
            experiment passed to it. **Use only if the experiment had just
            been saved, as otherwise experiment state could get corrupted.**
    """

    experiment: Experiment
    generation_strategy: GenerationStrategy
    # pyre-fixme[24]: Generic type `LoggerAdapter` expects 1 type parameter.
    logger: LoggerAdapter
    # Mapping of form {short string identifier -> message to show in reported
    # results}. This is a mapping and not a list to allow for changing of
    # some optimization messages throughout the course of the optimization
    # (e.g. progress report of the optimization).
    markdown_messages: dict[str, MessageOutput]

    # Number of trials that existed on the Orchestrator's experiment before
    # the Orchestrator instantiation with that experiment.
    _num_preexisting_trials: int
    # Number of trials remaining to be scheduled during run_trials_and_yield_results.
    # Saved as a property so that it can be accessed after optimization is complex (ex.
    # for global stopping saving calculation).
    _num_remaining_requested_trials: int = 0
    # Total number of MetricFetchEs encountered during the course of optimization. Note
    # this is different from and may be greater than the number of trials that have
    # been marked either FAILED or ABANDONED due to metric fetching errors.
    _num_metric_fetch_e_encountered: int = 0
    # Number of trials that have been marked either FAILED or ABANDONED due to
    # MetricFetchE being encountered during _fetch_and_process_trials_data_results
    _num_trials_bad_due_to_err: int = 0
    # Keeps track of whether the allowed failure rate has been exceeded during
    # the optimization. If true, allows any pending trials to finish and raises
    # an error through self._complete_optimization.
    _failure_rate_has_been_exceeded: bool = False
    # Timestamp of last optimization start time (milliseconds since Unix epoch);
    # recorded in each `run_n_trials`.
    _latest_optimization_start_timestamp: int | None = None
    # Timeout setting for current optimization.
    _timeout_hours: float | None = None
    # Timestamp of when the last deployed trial started running.
    _latest_trial_start_timestamp: float | None = None
    # Will be set to `True` if generation strategy signals that the optimization
    # is complete, in which case the optimization should gracefully exit early.
    _optimization_complete: bool = False
    # This will disable the global stopping strategy. It is useful in some
    # applications where the user wants to run the optimization loop to exhaust
    # the declared number of trials.
    __ignore_global_stopping_strategy: bool = False
    # Default kwargs when fetching data if not overridden on `OrchestratorOptions`
    DEFAULT_FETCH_KWARGS = {
        "overwrite_existing_data": True,
    }

    def __init__(
        self,
        experiment: Experiment,
        generation_strategy: GenerationStrategy,
        options: OrchestratorOptions,
        db_settings: DBSettings | None = None,
        _skip_experiment_save: bool = False,
    ) -> None:
        self.experiment = experiment
        # Set up logger with an optional filepath handler. Note: we set the
        # logger before setting options since that can trigger errors.
        self._set_logger(options=options)
        self.options = options
        # NOTE: Parallelism schedule is embedded in the generation
        # strategy, as `GenerationStep.max_parallelism`.
        self.generation_strategy = generation_strategy

        if not isinstance(experiment, Experiment):
            raise TypeError(f"{experiment} is not an Ax experiment.")
        if not isinstance(generation_strategy, GenerationStrategy):
            raise TypeError(f"{generation_strategy} is not a generation strategy.")

        # Initialize storage layer for the orchestrator.
        super().__init__(
            db_settings=db_settings,
            logging_level=self.options.logging_level,
            suppress_all_errors=self.options.suppress_storage_errors_after_retries,
        )

        # Validate experiment and GS; ensure that experiment has immutable
        # search space and opt. config to avoid storing their  copies on each
        # generator run.
        self._validate_remaining_trials(experiment=experiment)
        if self.options.enforce_immutable_search_space_and_opt_config:
            self._enforce_immutable_search_space_and_opt_config()

        if self.db_settings_set and not _skip_experiment_save:
            self._maybe_save_experiment_and_generation_strategy(
                experiment=experiment, generation_strategy=generation_strategy
            )

        # Number of trials that existed on experiment before this orchestrator.
        self._num_preexisting_trials = len(experiment.trials)
        # Whether to log the reason why no trials were generated next time
        # we prepare new trials for deployment. Used to avoid spamming logs
        # when trials are not generated for the same reason multiple times in
        # a row.
        self._log_next_no_trials_reason = True
        self.markdown_messages = {
            "Generation strategy": MessageOutput(
                text=GS_TYPE_MSG.format(gs_name=generation_strategy.name),
                priority=OutputPriority.DEBUG,
            ),
        }

    @classmethod
    def get_default_db_settings(cls) -> DBSettings:
        raise NotImplementedError(
            "Base `Orchestrator` does not specify default `DBSettings`. "
            "DBSettings are required to leverage SQL storage functionality "
            "and can be specified as argument to `Orchestrator` constructor or "
            "via `get_default_db_settings` implementation on given orchestrator."
        )

    @classmethod
    def from_stored_experiment(
        cls,
        experiment_name: str,
        options: OrchestratorOptions,
        db_settings: DBSettings | None = None,
        generation_strategy: GenerationStrategy | None = None,
        reduced_state: bool = True,
        **kwargs: Any,
    ) -> Orchestrator:
        """Create a ``Orchestrator`` with a previously stored experiment, which
        the Orchestrator should resume.

        Args:
            experiment_name: Experiment to load and resume.
            options: ``OrchestratorOptions``, with which to set up the new orchestrator.
            db_settings: Optional ``DBSettings``, which to use for reloading the
                experiment; also passed as ``db_settings`` argument to the
                Orchestrator constructor.
            generation_strategy: Generation strategy to use to provide candidates
                for the resumed optimization. Provide this argument only if
                the experiment does not already have a generation strategy
                associated with it.
            kwargs: Kwargs to pass through to the ``Orchestrator`` constructor.
        """
        dbs = WithDBSettingsBase(
            db_settings=db_settings or cls.get_default_db_settings()
        )
        exp, gs = dbs._load_experiment_and_generation_strategy(
            experiment_name=experiment_name,
            reduced_state=reduced_state,
        )
        if db_settings:
            kwargs = {**kwargs, "db_settings": db_settings}
        if not exp:
            raise ValueError(f"Experiment {experiment_name} not found.")

        if not gs and not generation_strategy:
            raise ValueError(
                f"Experiment {experiment_name} did not have a generation "
                "strategy associated with it in the database, so a new "
                "generation strategy must be provided as argument to "
                "`Orchestrator.from_stored_experiment`."
            )

        if gs and generation_strategy and gs != generation_strategy:
            # NOTE: In the future we may want to allow overriding of GS,
            # in which case we can add a flag to this function and allow
            # the override with warning.
            raise UnsupportedError(
                "Experiment was associated with generation strategy "
                f"{gs.name} in DB, but a new generation strategy "
                f"{generation_strategy.name} was provided. To use "
                "the generation strategy currently in DB, do not "
                "specify the `geneneration_strategy` kwarg."
            )

        orchestrator = cls(
            experiment=exp,
            generation_strategy=none_throws(generation_strategy or gs),
            options=options,
            # No need to resave the experiment we just reloaded.
            _skip_experiment_save=True,
            # NOTE: `kwargs` can include `db_settings` if those were
            # provided to this function.
            **kwargs,
        )
        return orchestrator

    @property
    def options(self) -> OrchestratorOptions:
        """Orchestrator options."""
        return self._options  # pyre-ignore [16]

    @options.setter
    def options(self, options: OrchestratorOptions) -> None:
        """Set Orchestrator options."""
        self._validate_options(options=options)
        self._options = options
        # validate runners and metrics since validate_metrics is an option
        self._validate_runner_and_implemented_metrics(experiment=self.experiment)

    @property
    def trial_type(self) -> str | None:
        """Trial type for the experiment this Orchestrator is running.

        This returns None if the experiment is not a MultitypeExperiment

        Returns:
            Trial type for the experiment this Orchestrator is running if the
            experiment is a MultiTypeExperiment and None otherwise.
        """
        if isinstance(self.experiment, MultiTypeExperiment):
            return self.options.mt_experiment_trial_type
        return None

    @property
    def running_trials(self) -> list[BaseTrial]:
        """Currently running trials.

        Note: if the experiment is a MultiTypeExperiment, then this will
        only fetch trials of type `orchestrator.trial_type`.


        Returns:
            List of trials that are currently running.
        """
        return filter_trials_by_type(
            trials=self.experiment.trials_by_status[TrialStatus.RUNNING],
            trial_type=self.trial_type,
        )

    @property
    def trials(self) -> list[BaseTrial]:
        """All trials.

        Note: if the experiment is a MultiTypeExperiment, then this will
        only fetch trials of type `orchestrator.trial_type`.

        Returns:
            List of trials that are currently running.
        """
        return filter_trials_by_type(
            trials=list(self.experiment.trials.values()), trial_type=self.trial_type
        )

    @property
    def running_trial_indices(self) -> set[int]:
        """Currently running trials.

        Returns:
            List of trials that are currently running.
        """
        return get_trial_indices_for_statuses(
            experiment=self.experiment,
            statuses={TrialStatus.RUNNING},
            trial_type=self.trial_type,
        )

    @property
    def failed_abandoned_trial_indices(self) -> set[int]:
        """Failed or abandoned trials.

        Note: if the experiment is a MultiTypeExperiment, then this will
        only fetch trials of type `orchestrator.trial_type`.

        Returns:
            List of trials that are currently running.
        """
        return get_trial_indices_for_statuses(
            experiment=self.experiment,
            statuses={TrialStatus.ABANDONED, TrialStatus.FAILED},
            trial_type=self.trial_type,
        )

    @property
    def pending_trials(self) -> list[BaseTrial]:
        """Running or staged trials on the experiment this Orchestrator is
        running.

        Note: if the experiment is a MultiTypeExperiment, then this will
        only fetch trials of type `orchestrator.trial_type`.

        Returns:
            List of trials that are currently running or staged.
        """
        staged_trials = filter_trials_by_type(
            trials=self.experiment.trials_by_status[TrialStatus.STAGED],
            trial_type=self.trial_type,
        )
        return self.running_trials + staged_trials

    @property
    def candidate_trials(self) -> list[BaseTrial]:
        """Candidate trials on the experiment this Orchestrator is running.

        Note: if the experiment is a MultiTypeExperiment, then this will
        only fetch trials of type `orchestrator.trial_type`.

        Returns:
            List of trials that are currently candidates.
        """
        return filter_trials_by_type(
            trials=self.experiment.trials_by_status[TrialStatus.CANDIDATE],
            trial_type=self.trial_type,
        )

    @property
    def trials_expecting_data(self) -> list[BaseTrial]:
        """Trials expecting data.

        Note: if the experiment is a MultiTypeExperiment, then this will
        only fetch trials of type `orchestrator.trial_type`.
        """
        trials = []
        for trial in self.experiment.trials.values():
            if trial.status.expecting_data:
                if self.trial_type is None or trial.trial_type == self.trial_type:
                    trials.append(trial)
        return trials

    @property
    def runner(self) -> Runner:
        """``Runner`` specified on the experiment associated with this ``Orchestrator``
        instance.
        """
        if self.trial_type is not None:
            runner = assert_is_instance(
                self.experiment, MultiTypeExperiment
            ).runner_for_trial_type(trial_type=none_throws(self.trial_type))
        else:
            runner = self.experiment.runner
        if runner is None:
            raise UnsupportedError(
                "`Orchestrator` requires that experiment specifies a `Runner`."
            )
        return runner

    def __repr__(self) -> str:
        """Short user-friendly string representation."""
        if not hasattr(self, "experiment"):
            # Experiment, generation strategy, etc. attributes have not
            # yet been set.
            return f"{self.__class__.__name__}"
        return (
            f"{self.__class__.__name__}(experiment={self.experiment}, "
            f"generation_strategy={self.generation_strategy}, options="
            f"{self.options})"
        )

    # ---------- Methods below should generally not be modified in subclasses! ---------
    # ---------- I. Methods that are often called outside the `Orchestrator`. ---------

    def generate_candidates(
        self,
        num_trials: int = 1,
        reduce_state_generator_runs: bool = False,
        remove_stale_candidates: bool = False,
    ) -> tuple[list[BaseTrial], Exception | None]:
        """Fetch the latest data and generate new candidate trials.

        Args:
            num_trials: Number of candidate trials to generate.
            reduce_state_generator_runs: Flag to determine
                whether to save model state for every generator run (default)
                or to only save model state on the final generator run of each
                batch.
            remove_stale_candidates: If true, mark any existing candidate trials
                failed before trial generation because:
                - they should not be treated as pending points
                - they will no longer be relevant

        Returns:
            List of trials, empty if generation is not possible.
        """
        if remove_stale_candidates:
            stale_candidate_trials = self.experiment.trials_by_status[
                TrialStatus.CANDIDATE
            ]
            self.logger.debug(
                "Marking the following trials as failed because they are stale: "
                f"{[t.index for t in stale_candidate_trials]}"
            )
            for trial in stale_candidate_trials:
                trial.mark_failed(reason="Newer candidates generated.", unsafe=True)
        else:
            stale_candidate_trials = []
        new_trials, err = self._get_next_trials(
            num_trials=num_trials,
            n=self.options.batch_size,
        )
        if len(new_trials) > 0:
            new_generator_runs = [gr for t in new_trials for gr in t.generator_runs]
            self._save_or_update_trials_and_generation_strategy_if_possible(
                experiment=self.experiment,
                trials=new_trials + stale_candidate_trials,
                generation_strategy=self.generation_strategy,
                new_generator_runs=new_generator_runs,
                reduce_state_generator_runs=reduce_state_generator_runs,
            )
        return new_trials, err

    def run_n_trials(
        self,
        max_trials: int,
        ignore_global_stopping_strategy: bool = False,
        timeout_hours: float | None = None,
        idle_callback: Callable[[Orchestrator], None] | None = None,
    ) -> OptimizationResult:
        """Run up to ``max_trials`` trials; will run all ``max_trials`` unless
        completion criterion is reached. For base ``Orchestrator``, completion criterion
        is reaching total number of trials set in ``OrchestratorOptions``, so if that
        option is not specified, this function will run exactly ``max_trials`` trials
        always.

        Args:
            max_trials: Maximum number of trials to run.
            ignore_global_stopping_strategy: If set, Orchestrator will skip the global
                stopping strategy in ``should_consider_optimization_complete``.
            timeout_hours: Limit on length of ths optimization; if reached, the
                optimization will abort even if completon criterion is not yet reached.
            idle_callback: Callable that takes a Orchestrator instance as an argument to
                deliver information while the trials are still running. Any output of
                `idle_callback` will not be returned, so `idle_callback` must expose
                information in some other way. For example, it could print something
                about the state of the Orchestrator or underlying experiment to STDOUT,
                write something to a database, or modify a Plotly figure or other object
                in place. `ax.service.utils.report_utils.get_figure_and_callback` is a
                helper function for generating a callback that will update a Plotly
                figure.

        Example:
            >>> trials_info = {"n_completed": None}
            >>>
            >>> def write_n_trials(orchestrator:Orchestrator) -> None:
            ...     trials_info["n_completed"] = len(orchestrator.experiment.trials)
            >>>
            >>> orchestrator.run_n_trials(
            ...     max_trials=3, idle_callback=write_n_trials
            ... )
            >>> print(trials_info["n_completed"])
            3
        """
        self.poll_and_process_results()
        for _ in self.run_trials_and_yield_results(
            max_trials=max_trials,
            ignore_global_stopping_strategy=ignore_global_stopping_strategy,
            timeout_hours=timeout_hours,
            idle_callback=idle_callback,
        ):
            pass
        return self.summarize_final_result()

    def run_all_trials(
        self,
        timeout_hours: float | None = None,
        idle_callback: Callable[[Orchestrator], None] | None = None,
    ) -> OptimizationResult:
        """Run all trials until ``should_consider_optimization_complete`` yields
        true (by default, ``should_consider_optimization_complete`` will yield true when
        reaching the ``num_trials`` setting, passed to Orchestrator on instantiation as
        part of ``OrchestratorOptions``).

        NOTE: This function is available only when ``OrchestratorOptions.num_trials`` is
        specified.

        Args:
            timeout_hours: Limit on length of ths optimization; if reached, the
                optimization will abort even if completon criterion is not yet reached.
            idle_callback: Callable that takes a Orchestrator instance as an argument to
                deliver information while the trials are still running. Any output of
                `idle_callback` will not be returned, so `idle_callback` must expose
                information in some other way. For example, it could print something
                about the state of the Orchestrator or underlying experiment to STDOUT,
                write something to a database, or modify a Plotly figure or other object
                in place. `ax.service.utils.report_utils.get_figure_and_callback` is a
                helper function for generating a callback that will update a Plotly
                figure.

        Example:
            >>> trials_info = {"n_completed": None}
            >>>
            >>> def write_n_trials(orchestrator:Orchestrator) -> None:
            ...     trials_info["n_completed"] = len(orchestrator.experiment.trials)
            >>>
            >>> orchestrator.run_all_trials(
            ...     timeout_hours=0.1, idle_callback=write_n_trials
            ... )
            >>> print(trials_info["n_completed"])
        """
        if self.options.total_trials is None:
            # NOTE: Capping on number of trials will likely be needed as fallback
            # for most stopping criteria, so we ensure `num_trials` is specified.
            raise ValueError(
                "Please either specify `num_trials` in `OrchestratorOptions` input "
                "to the `Orchestrator` or use `run_n_trials` instead of "
                "`run_all_trials`."
            )
        return self.run_n_trials(
            max_trials=none_throws(self.options.total_trials),
            timeout_hours=timeout_hours,
            idle_callback=idle_callback,
        )

    def run_trials_and_yield_results(
        self,
        max_trials: int,
        ignore_global_stopping_strategy: bool = False,
        timeout_hours: float | None = None,
        idle_callback: Callable[[Orchestrator], None] | None = None,
    ) -> Generator[dict[str, Any], None, None]:
        """Make continuous calls to `run` and `process_results` to run up to
        ``max_trials`` trials, until completion criterion is reached. This is the 'main'
        method of a ``Orchestrator``.

        Args:
            max_trials: Maximum number of trials to run in this generator. The
                generator will run trials until a completion criterion is reached,
                a completion signal is received from the generation strategy, or
                ``max_trials`` trials have been run (whichever happens first).
            ignore_global_stopping_strategy: If set, Orchestrator will skip the global
                stopping strategy in ``should_consider_optimization_complete``.
            timeout_hours: Maximum number of hours, for which
                to run the optimization. This function will abort after running
                for `timeout_hours` even if stopping criterion has not been reached.
                If set to `None`, no optimization timeout will be applied.
            idle_callback: Callable that takes a Orchestrator instance as an argument to
                deliver information while the trials are still running. Any output of
                `idle_callback` will not be returned, so `idle_callback` must expose
                information in some other way. For example, it could print something
                about the state of the Orchestrator or underlying experiment to STDOUT,
                write something to a database, or modify a Plotly figure or other object
                in place. `ax.service.utils.report_utils.get_figure_and_callback` is a
                helper function for generating a callback that will update a Plotly
                figure.
        """
        if max_trials < 0:
            raise ValueError(f"Expected `max_trials` >= 0, got {max_trials}.")

        if timeout_hours is not None:
            if timeout_hours < 0:
                raise UserInputError(
                    f"Expected `timeout_hours` >= 0, got {timeout_hours}."
                )

        self._latest_optimization_start_timestamp = current_timestamp_in_millis()
        self.__ignore_global_stopping_strategy = ignore_global_stopping_strategy

        n_initial_candidate_trials = len(self.candidate_trials)
        if n_initial_candidate_trials == 0 and max_trials < 0:
            raise UserInputError(f"Expected `max_trials` >= 0, got {max_trials}.")

        # trials are pre-existing only if they do not still require running
        n_existing = len(self.trials) - n_initial_candidate_trials

        # Until completion criterion is reached or `max_trials` is scheduled,
        # schedule new trials and poll existing ones in a loop.
        self._num_remaining_requested_trials = max_trials
        while (
            self._num_remaining_requested_trials > 0
            and not self.should_consider_optimization_complete()[0]
        ):
            if self.should_abort_optimization(timeout_hours=timeout_hours):
                yield self._abort_optimization(num_preexisting_trials=n_existing)
                return

            # Run new trial evaluations until `run` returns `False`, which
            # means that there was a reason not to run more evaluations yet.
            # Also check that `max_trials` is not reached to not exceed it.
            n_remaining_to_generate = self._num_remaining_requested_trials - len(
                self.candidate_trials
            )
            while self._num_remaining_requested_trials > 0 and self.run(
                max_new_trials=n_remaining_to_generate,
                timeout_hours=timeout_hours,
            ):
                # Not checking `should_abort_optimization` on every trial for perf.
                # reasons.
                n_already_run_by_orchestrator = (
                    len(self.trials) - n_existing - len(self.candidate_trials)
                )
                self._num_remaining_requested_trials = (
                    max_trials - n_already_run_by_orchestrator
                )
                n_remaining_to_generate = self._num_remaining_requested_trials - len(
                    self.candidate_trials
                )
            # this is safeguard in case no trial statuses have been updated, and
            # wait_for_running_trials=False, in which case we do not want to continue
            # to loop and poll
            report_results = self._check_exit_status_and_report_results(
                n_existing=n_existing, idle_callback=idle_callback, force_refit=False
            )
            if report_results is None:
                return
            else:
                yield report_results

        # When done scheduling, wait for the remaining trials to finish running
        # (unless optimization is aborting, in which case stop right away).
        if self.running_trials:
            self.logger.info(
                "Done submitting trials, waiting for remaining "
                f"{len(self.running_trials)} running trials..."
            )

        while self.running_trials:
            if self.should_abort_optimization(timeout_hours=timeout_hours):
                yield self._abort_optimization(num_preexisting_trials=n_existing)
                return
            report_results = self._check_exit_status_and_report_results(
                n_existing=n_existing, idle_callback=idle_callback, force_refit=True
            )
            if report_results is None:
                return
            else:
                yield report_results

        yield self._complete_optimization(
            num_preexisting_trials=n_existing, idle_callback=idle_callback
        )
        return

    # -------- II. Methods that are typically called within the `Orchestrator`. -------

    @retry_on_exception(retries=3, no_retry_on_exception_types=NO_RETRY_EXCEPTIONS)
    def run_trials(self, trials: Iterable[BaseTrial]) -> dict[int, dict[str, Any]]:
        """Deployment function, runs a single evaluation for each of the
        given trials.

        Override default implementation on the ``Runner`` if its desirable to deploy
        trials in bulk.

        NOTE: the `retry_on_exception` decorator applied to this function should also
        be applied to its subclassing override if one is provided and retry behavior
        is desired.

        Args:
            trials: Iterable of trials to be deployed, each containing arms with
                parameterizations to be evaluated. Can be a ``Trial``
                if contains only one arm or a ``BatchTrial`` if contains
                multiple arms.

        Returns:
            Dict of trial index to the run metadata of that trial from the deployment
            process.
        """
        return self.runner.run_multiple(trials=trials)

    @retry_on_exception(retries=3, no_retry_on_exception_types=NO_RETRY_EXCEPTIONS)
    def poll_trial_status(
        self, poll_all_trial_statuses: bool = False
    ) -> dict[TrialStatus, set[int]]:
        """Polling function, checks the status of any non-terminal trials
        and returns their indices as a mapping from TrialStatus to a list of indices.

        NOTE: Does not need to handle waiting between polling while trials
        are running; that logic is handled in ``orchestrator.poll``, which calls
        this function.

        Returns:
            A dictionary mapping TrialStatus to a list of trial indices that have
            the respective status at the time of the polling. This does not need to
            include trials that at the time of polling already have a terminal
            (ABANDONED, FAILED, COMPLETED) status (but it may).
        """
        trials = (
            list(self.experiment.trials.values())
            if poll_all_trial_statuses
            else self.pending_trials
        )
        trials = filter_trials_by_type(trials=trials, trial_type=self.trial_type)
        if len(trials) == 0:
            return {}
        return self.runner.poll_trial_status(trials=trials)

    def wait_for_completed_trials_and_report_results(
        self,
        idle_callback: Callable[[Orchestrator], None] | None = None,
        force_refit: bool = False,
    ) -> dict[str, Any]:
        """Continuously poll for successful trials, with limited exponential
        backoff, and process the results. Stop once at least one successful
        trial has been found. This function can be overridden to a different
        waiting function as needed; it must call `poll_and_process_results`
        to ensure that trials that completed their evaluation are appropriately
        marked as 'COMPLETED' in Ax.

        Args:
            idle_callback: Callable that takes a Orchestrator instance as an argument to
                deliver information while the trials are still running. Any output of
                `idle_callback` will not be returned, so `idle_callback` must expose
                information in some other way. For example, it could print something
                about the state of the Orchestrator or underlying experiment to STDOUT,
                write something to a database, or modify a Plotly figure or other object
                in place. `ax.service.utils.report_utils.get_figure_and_callback` is a
                helper function for generating a callback that will update a Plotly
                figure.
            force_refit: Whether to force a refit of the model during report_results.

        Returns:
            Results of the optimization so far, represented as a
            dict. The contents of the dict depend on the implementation of
            `report_results` in the given `Orchestrator` subclass.
        """
        if self.options.init_seconds_between_polls is None:
            raise ValueError(
                "Default `wait_for_completed_trials_and_report_results` in base "
                "`Orchestrator` relies on non-null `init_seconds_between_polls` "
                "`Orchestrator` option."
            )

        seconds_between_polls = self.options.init_seconds_between_polls
        backoff_factor = self.options.seconds_between_polls_backoff_factor

        total_seconds_elapsed = 0
        while len(self.pending_trials) > 0 and not self.poll_and_process_results():
            if total_seconds_elapsed > MAX_SECONDS_BETWEEN_REPORTS:
                break  # If maximum wait time reached, check the stopping
                # criterion again and and re-attempt scheduling more trials.

            if idle_callback is not None:
                try:
                    idle_callback(self)
                except Exception as e:
                    self.logger.warning(
                        f"Exception raised in ``idle_callback``: {e}. "
                        "Continuing to poll for completed trials."
                    )

            log_seconds = (
                int(seconds_between_polls)
                if seconds_between_polls > 2
                else seconds_between_polls
            )
            self.logger.info(
                f"Waiting for completed trials (for {log_seconds} sec, "
                f"currently running trials: {len(self.running_trials)})."
            )

            sleep(seconds_between_polls)

            total_seconds_elapsed += seconds_between_polls
            seconds_between_polls *= backoff_factor

        if idle_callback is not None:
            idle_callback(self)
        return self.report_results(force_refit=force_refit)

    def should_consider_optimization_complete(self) -> tuple[bool, str]:
        """Whether this Orchestrator should consider this optimization complete and not
        run more trials (and conclude the optimization via ``_complete_optimization``).

        NOTE: An optimization is considered complete when a generation strategy signaled
        completion or when the ``should_consider_optimization_complete`` method on this
        Orchestrator evaluates to ``True``. ``should_consider_optimization_complete``
        method is also responsible for checking global_stopping_strategy's decision as
        well. Alongside the stop decision, this function returns a string describing the
        reason for stopping the optimization.
        """
        if self._optimization_complete:
            return True, ""
        if len(self.pending_trials) == 0 and self._get_max_pending_trials() == 0:
            return (
                True,
                "All pending trials have completed and max_pending_trials is zero.",
            )

        should_stop, message = self._should_stop_due_to_global_stopping_strategy()
        if not should_stop:
            if self.options.total_trials is None:
                return False, ""
            should_stop, message = self._should_stop_due_to_total_trials()

        if should_stop:
            self.logger.info(
                f"Completing the optimization: {message}. "
                f"`should_consider_optimization_complete` "
                f"is `True`, not running more trials."
            )
        return should_stop, message

    def should_abort_optimization(self, timeout_hours: float | None = None) -> bool:
        """Checks whether this Orchestrator has reached some intertuption / abort
        criterion, such as an overall optimization timeout, tolerated failure rate, etc.
        """
        # If failure rate has been exceeded, log a warning and make sure we are not
        # scheduling additional trials. Raises an exception after pending trials have
        # completed, but does not abort the optimization immediately.
        self.error_if_failure_rate_exceeded()

        # if optimization is timed out, return True, else return False
        latest_optimization_start_timestamp = self._latest_optimization_start_timestamp
        timeout_in_millis = (
            timeout_hours * 60 * 60 * 1000 if timeout_hours is not None else None
        )
        timed_out = False

        if (
            latest_optimization_start_timestamp is not None
            and timeout_in_millis is not None
        ):
            time_elapsed_in_millis = (
                current_timestamp_in_millis() - latest_optimization_start_timestamp
            )
            timed_out = time_elapsed_in_millis >= timeout_in_millis

        if timed_out:
            self.logger.error(
                "Optimization timed out (timeout hours: " f"{timeout_hours})!"
            )

        return timed_out

    def report_results(self, force_refit: bool = False) -> dict[str, Any]:
        """Optional user-defined function for reporting intermediate
        and final optimization results (e.g. make some API call, write to some
        other db). This function is called whenever new results are available during
        the optimization.

        Args:
            force_refit: Whether to force the implementation of this method to
                refit the model on generation strategy before using it to produce
                results to report (e.g. if using model to visualize data).

        Returns:
            An optional dictionary with any relevant data about optimization.
        """
        # TODO[T61776778]: add utility to get best trial from arbitrary exp.
        return {}

    def summarize_final_result(self) -> OptimizationResult:
        """Get some summary of result: which trial did best, what
        were the metric values, what were encountered failures, etc.
        """
        return OptimizationResult()

    def _check_if_failure_rate_exceeded(self, force_check: bool = False) -> bool:
        """Checks if the failure rate (set in Orchestrator options) has been exceeded at
        any point during the optimization.

        NOTE: Both FAILED and ABANDONED trial statuses count towards the failure rate.

        Args:
            force_check: Indicates whether to force a failure-rate check
                regardless of the number of trials that have been executed. If False
                (default), the check will be skipped if the optimization has fewer than
                five failed trials. If True, the check will be performed unless there
                are 0 failures.

        Effect on state:
            If the failure rate has been exceeded, a warning is logged and the private
            attribute `_failure_rate_has_been_exceeded` is set to True, which causes the
            `_get_max_pending_trials` to return zero, so that no further trials are
            scheduled and an error is raised at the end of the optimization.

        Returns:
            Boolean representing whether the failure rate has been exceeded.
        """
        if self._failure_rate_has_been_exceeded:
            return True

        num_bad_in_orchestrator = self._num_bad_in_orchestrator()
        # skip check if 0 failures
        if num_bad_in_orchestrator == 0:
            return False

        # skip check if fewer than min_failed_trials_for_failure_rate_check failures
        # unless force_check is True
        if (
            num_bad_in_orchestrator
            < self.options.min_failed_trials_for_failure_rate_check
            and not force_check
        ):
            return False

        num_ran_in_orchestrator = self._num_ran_in_orchestrator()
        failure_rate_exceeded = (
            num_bad_in_orchestrator / num_ran_in_orchestrator
        ) > self.options.tolerated_trial_failure_rate

        if failure_rate_exceeded:
            if self._num_trials_bad_due_to_err > num_bad_in_orchestrator / 2:
                self.logger.warning(
                    "MetricFetchE INFO: Sweep aborted due to an exceeded error rate, "
                    "which was primarily caused by failure to fetch metrics. Please "
                    "check if anything could cause your metrics to be flaky or "
                    "broken."
                )
            # NOTE: this private attribute causes `_get_max_pending_trials` to
            # return zero, which causes no further trials to be scheduled.
            self._failure_rate_has_been_exceeded = True
            return True

        return False

    def error_if_failure_rate_exceeded(self, force_check: bool = False) -> None:
        """Raises an exception if the failure rate (set in Orchestrator options) has
        been exceeded at any point during the optimization.

        NOTE: Both FAILED and ABANDONED trial statuses count towards the failure rate.

        Args:
            force_check: Indicates whether to force a failure-rate check
                regardless of the number of trials that have been executed. If False
                (default), the check will be skipped if the optimization has fewer than
                five failed trials. If True, the check will be performed unless there
                are 0 failures.
        """
        if self._check_if_failure_rate_exceeded(force_check=force_check):
            raise self._get_failure_rate_exceeded_error(
                num_bad_in_orchestrator=self._num_bad_in_orchestrator(),
                num_ran_in_orchestrator=self._num_ran_in_orchestrator(),
            )

    def _check_exit_status_and_report_results(
        self,
        n_existing: int,
        idle_callback: Callable[[Orchestrator], None] | None,
        force_refit: bool,
    ) -> dict[str, Any] | None:
        if not self.options.wait_for_running_trials:
            return None
        return self.wait_for_completed_trials_and_report_results(
            idle_callback, force_refit=True
        )

    def run(self, max_new_trials: int, timeout_hours: float | None = None) -> bool:
        """Schedules trial evaluation(s) if stopping criterion is not triggered,
        maximum parallelism is not currently reached, and capacity allows.
        Logs any failures / issues.

        Args:
            max_new_trials: Maximum number of new trials this function should generate
                and run (useful when generating and running trials in batches). Note
                that this function might also re-deploy existing ``CANDIDATE`` trials
                that failed to deploy before, which will not count against this number.
            timeout_hours: Maximum number of hours, for which
                to run the optimization. This function will abort after running
                for `timeout_hours` even if stopping criterion has not been reached.
                If set to `None`, no optimization timeout will be applied.

        Returns:
            Boolean representing success status.
        """
        (
            optimization_complete,
            completion_message,
        ) = self.should_consider_optimization_complete()
        if optimization_complete:
            return False

        if self.should_abort_optimization(timeout_hours=timeout_hours):
            self.logger.info(
                "`should_abort_optimization` is `True`, not running more trials."
            )
            return False

        # Check if capacity allows for running new evaluations and generate as many
        # trials as possible, limited by capacity and model requirements.
        self._sleep_if_too_early_to_poll()
        existing_trials, new_trials = self._prepare_trials(
            max_new_trials=max_new_trials
        )

        if not existing_trials and not new_trials:
            # Unable to gen. new run due to max parallelism limit or need for data
            # or unable to run trials due to lack of capacity.
            if self._optimization_complete:
                return False

            if len(self.pending_trials) < 1:
                raise OrchestratorInternalError(
                    "No trials are running but model requires more data. This is an "
                    "invalid state of the Orchestrator, as no more trials can be added "
                    "but also no more data is expected as there are no running trials. "
                    "This should be investigated."
                )
            self._log_next_no_trials_reason = False
            return False  # Nothing to run.

        if existing_trials:
            idcs = sorted(t.index for t in existing_trials)
            self.logger.debug(f"Will run pre-existing candidate trials: {idcs}.")

        all_trials = [*existing_trials, *new_trials]
        idcs_str = make_indices_str(indices=(t.index for t in all_trials))
        self.logger.info(f"Running trials {idcs_str}...")
        # TODO: Add optional timeout between retries of `run_trial(s)`.
        metadata = self.run_trials(trials=all_trials)
        self.logger.debug(f"Ran trials {idcs_str}.")
        if self.options.debug_log_run_metadata:
            self.logger.debug(f"Run metadata: {metadata}.")
        self._latest_trial_start_timestamp = current_timestamp_in_millis()
        self._update_and_save_trials(
            existing_trials=existing_trials, new_trials=new_trials, metadata=metadata
        )
        self._log_next_no_trials_reason = True
        return True

    def poll_and_process_results(self, poll_all_trial_statuses: bool = False) -> bool:
        """Takes the following actions:
            1. Poll trial runs for their statuses
            2. Find trials to fetch data for
            3. Apply new trial statuses
            4. Fetch data
            5. Early-stop trials where possible
            6. Save modified trials, having either new statuses or new data

        Returns:
            A boolean representing whether any trial evaluations completed
            or have been marked as failed or abandoned, changing the number of
            currently running trials.
        """
        self._sleep_if_too_early_to_poll()

        # POLL TRIAL STATUSES
        new_status_to_trial_idcs = self.poll_trial_status(
            poll_all_trial_statuses=poll_all_trial_statuses
        )

        trial_indices_with_updated_data_or_status = set()

        # GET TRIALS TO FETCH DATA FOR
        # This must be done before updating the trial statuses, so we can differentiate
        # newly and previously completed trials.
        trial_indices_to_fetch = self._get_trial_indices_to_fetch(
            new_status_to_trial_idcs=new_status_to_trial_idcs
        )

        # UPDATE TRIAL STATUSES
        trial_indices_with_updated_statuses = self._apply_new_trial_statuses(
            new_status_to_trial_idcs=new_status_to_trial_idcs,
        )
        updated_any_trial_status = len(trial_indices_with_updated_statuses) > 0
        trial_indices_with_updated_data_or_status.update(
            trial_indices_with_updated_statuses
        )

        # FETCH DATA FOR TRIALS EXPECTING DATA
        trial_indices_with_new_data = (
            self._fetch_data_and_return_trial_indices_with_new_data(
                trial_idcs=trial_indices_to_fetch,
            )
        )
        trial_indices_with_updated_data_or_status.update(trial_indices_with_new_data)

        # EARLY STOP TRIALS
        stop_trial_info = early_stopping_utils.should_stop_trials_early(
            early_stopping_strategy=self.options.early_stopping_strategy,
            trial_indices=self.running_trial_indices,
            experiment=self.experiment,
            current_node=self.generation_strategy._curr,
        )
        self.experiment.stop_trial_runs(
            trials=[self.experiment.trials[trial_idx] for trial_idx in stop_trial_info],
            reasons=list(stop_trial_info.values()),
        )
        if len(stop_trial_info) > 0:
            trial_indices_with_updated_data_or_status.update(set(stop_trial_info))
            updated_any_trial_status = True

        # UPDATE TRIALS IN DB
        if (
            len(trial_indices_with_updated_data_or_status) > 0
        ):  # Only save if there were updates.
            self.logger.debug(
                f"Updating {len(trial_indices_with_updated_data_or_status)} "
                "trials in DB."
            )
            self._save_or_update_trials_in_db_if_possible(
                experiment=self.experiment,
                trials=[
                    self.experiment.trials[i]
                    for i in trial_indices_with_updated_data_or_status
                ],
            )

        return updated_any_trial_status

    # ------------------------- III. Protected helpers. -----------------------

    def _fetch_data_and_return_trial_indices_with_new_data(
        self, trial_idcs: set[int]
    ) -> set[int]:
        """Fetch data for any trials on the experiment that are expecting new data.

        Args:
            trial_idcs: A set of trial indices to fetch data for.

        Returns:
            Set of trial indices that were updated with new data.  We're not asserting
            that the new data is different than the old data, but may want to
            in the future.
        """
        if len(trial_idcs) > 0:
            results = self._fetch_and_process_trials_data_results(
                trial_indices=trial_idcs,
            )
            return {
                i
                for i, results_by_metric_name in results.items()
                for r in results_by_metric_name.values()
                if r.is_ok()
            }
        return set()

    def _num_bad_in_orchestrator(self) -> int:
        """Returns the number of trials that have failed or been abandoned in the
        orchestrator.
        """
        # We only count failed trials with indices that came after the preexisting
        # trials on experiment before Orchestrator use.
        return sum(
            1
            for f in self.failed_abandoned_trial_indices
            if f >= self._num_preexisting_trials
        )

    def _num_ran_in_orchestrator(self) -> int:
        """Returns the number of trials that have been run by the orchestrator."""
        return len(self.experiment.trials) - self._num_preexisting_trials

    def _apply_new_trial_statuses(
        self, new_status_to_trial_idcs: dict[TrialStatus, set[int]]
    ) -> set[int]:
        """Apply new trial statuses to the experiment according to poll results.

        Args:
            new_status_to_trial_idcs: Changes to be applied to trial statuses from
                poll_trial_status.

        Returns:
            Set of trial indices that were updated with new statuses.
        """
        updated_trial_indices = set()
        for status, trial_idcs in new_status_to_trial_idcs.items():
            if status.is_candidate or status.is_deployed:
                # No need to consider candidate, staged or running trials here (none of
                # these trials should actually be candidates, but we can filter on that)
                continue

            if len(trial_idcs) > 0:
                idcs = make_indices_str(indices=trial_idcs)
                self.logger.info(f"Retrieved {status.name} trials: {idcs}.")

            # Update trial statuses and record which trials were updated.
            trials = self.experiment.get_trials_by_indices(trial_idcs)
            updated_trial_indices.update(trial_idcs)
            for trial in trials:
                if status.is_failed or status.is_abandoned:
                    try:
                        reason = self.runner.poll_exception(trial)
                        trial.mark_as(status=status, unsafe=True, reason=reason)
                    except NotImplementedError:
                        # Some runners do not implement poll_failure_reason, so
                        # we fall back to marking the without a reason.
                        trial.mark_as(status=status, unsafe=True)
                else:
                    trial.mark_as(status=status, unsafe=True)
        return updated_trial_indices

    def _identify_trial_indices_to_fetch(
        self,
        old_status_to_trial_idcs: Mapping[TrialStatus, set[int]],
        new_status_to_trial_idcs: Mapping[TrialStatus, set[int]],
    ) -> set[int]:
        """
        Identify trial indices to fetch data for based on changes in trial statuses.

        Args:
            old_status_to_trial_idcs: Mapping of old trial statuses
                to their corresponding trial indices.
            new_status_to_trial_idcs: Mapping of new trial statuses
                to their corresponding trial indices.
        Returns:
            Set of trial indices to fetch data for.
        """
        # Get newly completed trials
        prev_completed_trial_idcs = old_status_to_trial_idcs.get(
            TrialStatus.COMPLETED, set()
        ) | old_status_to_trial_idcs.get(TrialStatus.EARLY_STOPPED, set())

        newly_completed = (
            new_status_to_trial_idcs.get(TrialStatus.COMPLETED, set())
            - prev_completed_trial_idcs
        )

        idcs = make_indices_str(indices=newly_completed)
        if newly_completed:
            self.logger.debug(f"Will fetch data for newly completed trials: {idcs}.")
        else:
            self.logger.debug("No newly completed trials; not fetching data for any.")

        # Get running trials with metrics available while running
        running_trial_indices_with_metrics = set()
        if any(
            m.is_available_while_running() for m in self.experiment.metrics.values()
        ):
            running_trial_indices_with_metrics = new_status_to_trial_idcs.get(
                TrialStatus.RUNNING, set()
            ) | old_status_to_trial_idcs.get(TrialStatus.RUNNING, set())

            for status, indices in new_status_to_trial_idcs.items():
                if status.is_terminal and indices:
                    running_trial_indices_with_metrics -= indices

            if running_trial_indices_with_metrics:
                idcs = make_indices_str(indices=running_trial_indices_with_metrics)
                self.logger.debug(
                    f"Will fetch data for trials: {idcs} because some metrics "
                    "on experiment are available while trials are running."
                )

        # Get previously completed trials with new data after completion
        recently_completed_trial_indices = self._get_recently_completed_trial_indices()
        if len(recently_completed_trial_indices) > 0:
            idcs = make_indices_str(indices=recently_completed_trial_indices)
            self.logger.debug(
                f"Will fetch data for trials: {idcs} because some metrics "
                "on experiment have new data after completion."
            )

        # Combine all trial indices to fetch data for
        trial_indices_to_fetch = (
            newly_completed
            | running_trial_indices_with_metrics
            | recently_completed_trial_indices
        )

        return trial_indices_to_fetch

    def _get_trial_indices_to_fetch(
        self, new_status_to_trial_idcs: Mapping[TrialStatus, set[int]]
    ) -> set[int]:
        """Get trial indices to fetch data for the experiment given
        `new_status_to_trial_idcs` and metric properties.  This should include:
            - newly completed trials
            - running trials if the experiment has metrics available while running
            - previously completed (or early stopped) trials if the experiment
                has metrics with new data after completion which finished recently

        Args:
            new_status_to_trial_idcs: Changes about to be applied to trial statuses.

        Returns:
            Set of trial indices to fetch data for.
        """
        old_status_to_trial_idcs = {status: set() for status in TrialStatus}

        for trial in self.trials:
            old_status_to_trial_idcs[trial.status].add(trial.index)

        return self._identify_trial_indices_to_fetch(
            old_status_to_trial_idcs=old_status_to_trial_idcs,
            new_status_to_trial_idcs=new_status_to_trial_idcs,
        )

    def _get_recently_completed_trial_indices(self) -> set[int]:
        """Get trials that have completed within the max period specified by metrics."""
        if len(self.experiment.metrics) == 0:
            return set()

        max_period = max(
            m.period_of_new_data_after_trial_completion()
            for m in self.experiment.metrics.values()
        )
        return {
            t.index
            for t in self.trials_expecting_data
            if t.time_completed is not None
            and datetime.now() - none_throws(t.time_completed) < max_period
        }

    def _process_completed_trials(self, newly_completed: set[int]) -> None:
        # Fetch the data for newly completed trials; this will cache the data
        # for all metrics. By pre-caching the data now, we remove the need to
        # fetch it during candidate generation.
        idcs = make_indices_str(indices=newly_completed)
        self.logger.debug(f"Fetching data for trials: {idcs}.")
        self._fetch_and_process_trials_data_results(
            trial_indices=newly_completed,
        )

    def _abort_optimization(self, num_preexisting_trials: int) -> dict[str, Any]:
        """Conclude optimization without waiting for anymore running trials and
        return results so far via `report_results`.
        """
        self._record_optimization_complete_message()
        return self.report_results(force_refit=True)

    def _complete_optimization(
        self,
        num_preexisting_trials: int,
        idle_callback: Callable[[Orchestrator], None] | None = None,
    ) -> dict[str, Any]:
        """Conclude optimization with waiting for anymore running trials and
        return final results via `wait_for_completed_trials_and_report_results`.
        """
        self._record_optimization_complete_message()
        res = self.wait_for_completed_trials_and_report_results(
            idle_callback=idle_callback, force_refit=True
        )
        # Raise an error if the failure rate exceeds tolerance at the
        # end of the optimization.
        self.error_if_failure_rate_exceeded(force_check=True)
        self._warn_if_non_terminal_trials()
        return res

    def _validate_options(self, options: OrchestratorOptions) -> None:
        """Validates `OrchestratorOptions` for compatibility with given
        `Orchestrator` class.
        """
        if not (0.0 <= options.tolerated_trial_failure_rate < 1.0):
            raise ValueError("`tolerated_trial_failure_rate` must be in [0, 1).")

        if options.early_stopping_strategy is not None and options.validate_metrics:
            if not any(
                m.is_available_while_running() for m in self.experiment.metrics.values()
            ):
                raise ValueError(
                    "Can only specify an early stopping strategy if at least one "
                    "metric is marked as `is_available_while_running`. Otherwise, we "
                    "will be unable to fetch intermediate results with which to "
                    "evaluate early stopping criteria."
                )
        if isinstance(self.experiment, MultiTypeExperiment):
            if options.mt_experiment_trial_type is None:
                raise UserInputError(
                    "Must specify `mt_experiment_trial_type` for MultiTypeExperiment."
                )
            if not self.experiment.supports_trial_type(
                options.mt_experiment_trial_type
            ):
                raise ValueError(
                    "Experiment does not support trial type "
                    f"{options.mt_experiment_trial_type}."
                )
        elif options.mt_experiment_trial_type is not None:
            raise UserInputError(
                "`mt_experiment_trial_type` must be None unless the experiment is a "
                "MultiTypeExperiment."
            )

    def _get_max_pending_trials(self) -> int:
        """Returns the maximum number of pending trials specified in the options, or
        zero, if the failure rate limit has been exceeded at any point during the
        optimization.
        """
        if self._failure_rate_has_been_exceeded:
            return 0
        return self.options.max_pending_trials

    def _prepare_trials(
        self, max_new_trials: int
    ) -> tuple[list[BaseTrial], list[BaseTrial]]:
        """Prepares one trial or multiple trials for deployment, based on
        whether `run_trials_in_batches` is set to `True` in this Orchestrator's
        options.

        NOTE: If running trials in batches, exact number of trials run at once
        is determined by available capacity and generation strategy's
        requirement for more data and parallelism limitation.

        Args:
            max_new_trials: Maximum number of new trials to generate.

        Returns:
            Two lists of trials:
            - list of existing candidate trials whose deployment was attempted
              but failed before (empty if there were no such trials),
            - list of new candidate trials that were created in the course of
              this function (empty if no new trials were generated).
        """
        # 1. Determine available capacity for running trials.
        capacity = self.runner.poll_available_capacity()
        if capacity != -1 and capacity < 1:  # -1 indicates unlimited capacity.
            self.logger.debug("There is no capacity to run any trials.")
            return [], []

        # 2. Determine actual number of trials to run based on capacity,
        # limit on pending trials and limit on total trials.
        n = capacity if self.options.run_trials_in_batches else 1
        total_trials = self.options.total_trials
        max_pending_trials = self._get_max_pending_trials()

        num_pending_trials = len(self.pending_trials)
        max_pending_upper_bound = max_pending_trials - num_pending_trials
        if max_pending_upper_bound < 1:
            self.logger.debug(
                f"`max_pending_trials={max_pending_trials}` and {num_pending_trials} "
                "trials are currently pending; not initiating any additional trials."
            )
            return [], []
        n = max_pending_upper_bound if n == -1 else min(max_pending_upper_bound, n)

        if total_trials is not None:
            left_in_total = total_trials - len(self.trials_expecting_data)
            n = min(n, left_in_total)

        existing_candidate_trials = self.candidate_trials[:n]
        n_new = min(n - len(existing_candidate_trials), max_new_trials)
        new_trials, _err = (
            self._get_next_trials(num_trials=n_new, n=self.options.batch_size)
            if n_new > 0
            else (
                [],
                None,
            )
        )
        return existing_candidate_trials, new_trials

    def _get_next_trials(
        self, num_trials: int = 1, n: int | None = None
    ) -> tuple[list[BaseTrial], Exception | None]:
        """Produce up to `num_trials` new generator runs from the underlying
        generation strategy and create new trials with them. Logs errors
        encountered during generation.

        NOTE: Fewer than `num_trials` trials may be produced if generation
        strategy runs into its parallelism limit or needs more data to proceed.

        Returns:
            List of trials, empty if generation is not possible.
        """
        try:
            generator_runs = self._gen_new_trials_from_generation_strategy(
                num_trials=num_trials, n=n
            )
        except OptimizationComplete as err:
            completion_str = f"Optimization complete: {err}"
            self.logger.info(completion_str)
            self.markdown_messages["Optimization complete"] = MessageOutput(
                text=completion_str,
                priority=OutputPriority.DEBUG,
            )
            self._optimization_complete = True
            return [], err
        except DataRequiredError as err:
            # TODO[T62606107]: consider adding a `more_data_required` property to
            # check to generation strategy to avoid running into this exception.
            if self._log_next_no_trials_reason:
                self.logger.info(
                    "Generated all trials that can be generated currently. "
                    "Model requires more data to generate more trials."
                )
            self.logger.debug(f"Message from generation strategy: {err}")
            return [], err
        except MaxParallelismReachedException as err:
            # TODO[T62606107]: consider adding a `step_max_parallelism_reached`
            # check to generation strategy to avoid running into this exception.
            if self._log_next_no_trials_reason:
                self.logger.info(
                    "Generated all trials that can be generated currently. "
                    "Max parallelism currently reached."
                )
            self.logger.debug(f"Message from generation strategy: {err}")
            return [], err
        except AxGenerationException as err:
            if self._log_next_no_trials_reason:
                self.logger.info(
                    "Generated all trials that can be generated currently. "
                    "`generation_strategy` encountered an error "
                    f"{err}."
                )
            self.logger.debug(f"Message from generation strategy: {err}")
            return [], err
        except OptimizationConfigRequired as err:
            if self._log_next_no_trials_reason:
                self.logger.info(
                    "Generated all trials that can be generated currently. "
                    "`generation_strategy` requires an optimization config "
                    "to be set before generating more trials."
                )
            self.logger.debug(f"Message from generation strategy: {err}")
            return [], err

        if self.options.trial_type == TrialType.TRIAL and any(
            len(generator_run_list[0].arms) > 1 or len(generator_run_list) > 1
            for generator_run_list in generator_runs
        ):
            raise OrchestratorInternalError(
                "Generation strategy produced multiple arms when only one was expected."
            )
        trials = []
        for generator_run_list in generator_runs:
            if self.options.trial_type == TrialType.BATCH_TRIAL:
                trial = self.experiment.new_batch_trial(
                    generator_runs=generator_run_list,
                    ttl_seconds=self.options.ttl_seconds_for_trials,
                    trial_type=self.trial_type,
                )
                if self.options.status_quo_weight > 0:
                    trial.add_status_quo_arm(
                        weight=self.options.status_quo_weight,
                    )
            else:
                trial = self.experiment.new_trial(
                    generator_run=generator_run_list[0],
                    ttl_seconds=self.options.ttl_seconds_for_trials,
                    trial_type=self.trial_type,
                )

            trials.append(trial)
        return trials, None

    def _gen_new_trials_from_generation_strategy(
        self,
        num_trials: int,
        n: int | None = None,
    ) -> list[list[GeneratorRun]]:
        """Generates a list ``GeneratorRun``s of length of ``num_trials`` using the
        ``_gen_multiple`` method of the Orchestrator's ``generation_strategy``, taking
        into account any ``pending`` observations.
        """
        self.generation_strategy.experiment = self.experiment
        # For ``BatchTrial`-s, we generate trials using the new method that can
        # produce GRs for multiple trials, with multiple nodes. But we don't yet
        # want to enable that functionality for single-arm use cases of the
        # ``Orchestrator``, as it's still in development.
        if self.options.trial_type == TrialType.BATCH_TRIAL:
            grs = self.generation_strategy.gen(
                experiment=self.experiment,
                num_trials=num_trials,
                n=n,
            )
            return grs
        else:
            assert self.options.trial_type == TrialType.TRIAL  # Sanity check.
            pending = get_pending_observation_features_based_on_trial_status(
                experiment=self.experiment
            )
            grs = self.generation_strategy.gen(
                experiment=self.experiment,
                num_trials=num_trials,
                n=1,
                pending_observations=pending,
                fixed_features=get_fixed_features_from_experiment(
                    experiment=self.experiment
                ),
            )
            return grs
        # TODO: pass self.trial_type to GS.gen for multi-type experiments

    def _update_and_save_trials(
        self,
        existing_trials: list[BaseTrial],
        new_trials: list[BaseTrial],
        metadata: dict[int, dict[str, Any]],
        reduce_state_generator_runs: bool = False,
    ) -> None:
        """Updates trials with new run metadata and status; saves updates to DB.

        Args:
            exiting_trials: Trials that existed on this experiment during the
                previous call to this function (these are trials, deployment of
                which has already been attempted but failed, so we are
                re-attempting it; these trials are already saved in DB if using
                storage functionality).
            new_trials: Trials that were newly created (these trials are not
                yet saved in the DB if using storage functionality).
            metadata: Run metadata for the trials, from `orchestrator.run_trials`.
                Format is {trial index -> trial run metadata}. Trials present in
                the metadata dict will be considered `RUNNING`, and the rest of
                trials in `existing_trials` or `new_trials` (that are not present
                in `metadata`) will be left as `CANDIDATE`.
            reduce_state_generator_runs: Flag to determine
                whether to save model state for every generator run (default)
                or to only save model state on the final generator run of each
                batch.
        """

        # pyre-fixme[3]: Return type must be annotated.
        # pyre-fixme[2]: Parameter must be annotated.
        def _process_trial(trial):
            if trial.index in metadata:
                trial.update_run_metadata(metadata=metadata[trial.index])
                try:
                    trial.mark_running(no_runner_required=True)
                except ValueError as e:
                    self.logger.warning(
                        "Unable to mark trial as RUNNING due to the following error:\n"
                        + str(e)
                    )
            else:
                self.logger.debug(
                    f"Trial {trial.index} did not deploy, status: {trial.status}."
                )

        new_generator_runs = []
        for trial in existing_trials:
            _process_trial(trial)
        for trial in new_trials:
            new_generator_runs.extend(trial.generator_runs)
            _process_trial(trial)

        self._save_or_update_trials_and_generation_strategy_if_possible(
            experiment=self.experiment,
            trials=[*existing_trials, *new_trials],
            generation_strategy=self.generation_strategy,
            new_generator_runs=new_generator_runs,
            reduce_state_generator_runs=reduce_state_generator_runs,
        )

    def _sleep_if_too_early_to_poll(self) -> None:
        """Wait to query for capacity unless there has been enough time since last
        scheduling.
        """
        if self._latest_trial_start_timestamp is not None:
            seconds_since_run_trial = round(
                (
                    current_timestamp_in_millis()
                    - none_throws(self._latest_trial_start_timestamp)
                )
                / 1000
            )

            if seconds_since_run_trial < self.options.min_seconds_before_poll:
                sleep_duration_seconds = (
                    self.options.min_seconds_before_poll - seconds_since_run_trial
                )
                self.logger.debug(
                    f"Too early to poll, sleeping for {sleep_duration_seconds} seconds"
                )
                sleep(sleep_duration_seconds)

    def _set_logger(self, options: OrchestratorOptions) -> None:
        """Set up the logger with appropriate logging levels."""
        cls_name = self.__class__.__name__
        logger = get_logger(name=f"{__name__}.{cls_name}@{hex(id(self))}")
        set_ax_logger_levels(level=options.logging_level)
        if options.log_filepath is not None:
            handler = build_file_handler(
                filepath=none_throws(options.log_filepath),
                level=options.logging_level,
            )
            logger.addHandler(handler)
        self.logger = LoggerAdapter(logger, extra={"output_name": cls_name})

    def _validate_remaining_trials(self, experiment: Experiment) -> None:
        """Check how many trials are remaining in `total_trials` given the trials
        already on experiment and make sure that there will be trials for the
        Orchestrator to run.
        """
        if not experiment.trials or not self.options.total_trials:
            return

        total_trials = none_throws(self.options.total_trials)
        preexisting = len(experiment.trials)
        msg = (
            f"{experiment} already has {preexisting} trials associated with it. "
            f"Total trials setting for this Orchestrator is {total_trials}, so "
        )
        if preexisting >= total_trials:
            self.logger.warning(
                msg + "no more trials would be run in this Orchestrator if "
                "`orchestrator.run_all_trials` is called (but you can still use "
                "`orchestrator.run_n_trials` to run a fixed number of trials)."
            )
        else:
            self.logger.info(
                msg + "number of trials ran by `orchestrator.run_all_trials` would be "
                f"{total_trials - preexisting}."
            )

    def _validate_runner_and_implemented_metrics(self, experiment: Experiment) -> None:
        """Ensure that the experiment specifies runner and metrics; check that metrics
        are not base ``Metric``-s, which do not implement fetching logic.
        """
        # this will raise an exception if no runner is set on the experiment
        self.runner
        metrics_are_invalid = False
        if not experiment.metrics:
            msg = "`Orchestrator` requires that `experiment.metrics` not be None."
            metrics_are_invalid = True
        else:
            msg = (
                "`Orchestrator` requires that experiment specifies metrics "
                "with implemented fetching logic."
            )
            base_metrics = {
                m_name for m_name, m in experiment.metrics.items() if type(m) is Metric
            }
            if base_metrics:
                msg += f" Metrics {base_metrics} do not implement fetching logic."
                metrics_are_invalid = True

        if metrics_are_invalid:
            if self.options.validate_metrics:
                raise UnsupportedError(msg)
            else:
                self.logger.warning(msg)

    def _enforce_immutable_search_space_and_opt_config(self) -> None:
        """Experiments with immutable search space and optimization config don't
        need to keep copies of those objects on each generator run in the experiment,
        resulting in large performance gain in storage layer. In `Orchestrator`, we
        force-set this immutability on `Experiment`, since Orchestrator experiments
        are typically not human-in-the-loop.
        """
        if self.experiment.immutable_search_space_and_opt_config:
            return

        self.logger.info(
            f"`Orchestrator` requires experiment to have immutable search "
            "space and optimization config. Setting property "
            f"{Keys.IMMUTABLE_SEARCH_SPACE_AND_OPT_CONF.value} "
            "to `True` on experiment."
        )
        self.experiment._properties[Keys.IMMUTABLE_SEARCH_SPACE_AND_OPT_CONF.value] = (
            True
        )

    def _record_optimization_complete_message(self) -> None:
        """Adds a simple optimization completion message to this Orchestrator's markdown
        messages.
        """
        completion_msg = OPTIMIZATION_COMPLETION_MSG.format(
            num_trials=len(self.experiment.trials),
            experiment_name=(
                self.experiment.name if self.experiment._name is not None else "unnamed"
            ),
        )
        if "Optimization complete" in self.markdown_messages:
            self.markdown_messages["Optimization complete"].append(text=completion_msg)
        else:
            self.markdown_messages["Optimization complete"] = MessageOutput(
                text=completion_msg,
                priority=OutputPriority.DEBUG,
            )

    def _fetch_and_process_trials_data_results(
        self,
        trial_indices: Iterable[int],
    ) -> dict[int, dict[str, MetricFetchResult]]:
        """
        Fetches results from experiment and modifies trial statuses depending on
        success or failure.
        """

        try:
            kwargs = deepcopy(self.options.fetch_kwargs)
            for k, v in self.DEFAULT_FETCH_KWARGS.items():
                kwargs.setdefault(k, v)
            if kwargs.get("overwrite_existing_data") and kwargs.get(
                "combine_with_last_data"
            ):
                # to avoid error https://fburl.com/code/ilix4okj
                kwargs["overwrite_existing_data"] = False
            if self.trial_type is not None:
                metrics = assert_is_instance(
                    self.experiment, MultiTypeExperiment
                ).metrics_for_trial_type(trial_type=none_throws(self.trial_type))
                kwargs["metrics"] = metrics
            results = self.experiment.fetch_trials_data_results(
                trial_indices=trial_indices,
                **kwargs,
            )
        except Exception as e:
            self.logger.exception(
                f"Failed to fetch data for trials {trial_indices} with error: {e}"
            )
            return {}

        for trial_index, results_by_metric_name in results.items():
            for metric_name, result in results_by_metric_name.items():
                # If the fetch call succeeded, continue.
                if result.is_ok():
                    continue

                # Log the Err so the user is aware that something has failed, even if
                # we do not do anything
                metric_fetch_e = result.unwrap_err()

                # If the metric is available while running just continue (we can try
                # again later).
                # NOTE: We don't need to report fetching errors in this case either
                metric = self.experiment.metrics[metric_name]
                status = self.experiment.trials[trial_index].status
                if (
                    metric.is_available_while_running()
                    and status == TrialStatus.RUNNING
                ):
                    self.logger.info(
                        f"MetricFetchE INFO: Because {metric_name} is "
                        f"available_while_running and trial {trial_index} is still "
                        "RUNNING continuing the experiment and retrying on next "
                        "poll..."
                    )
                    continue

                self.logger.error(
                    f"Failed to fetch {metric_name} for trial {trial_index} with "
                    f"status {status}, found {metric_fetch_e}."
                )
                self._num_metric_fetch_e_encountered += 1
                self._report_metric_fetch_e(
                    trial=self.experiment.trials[trial_index],
                    metric_name=metric_name,
                    metric_fetch_e=metric_fetch_e,
                )

                # If the fetch failure was for a metric in the optimization config (an
                # objective or constraint) mark the trial as failed
                optimization_config = self.experiment.optimization_config
                if (
                    optimization_config is not None
                    and metric_name in optimization_config.metrics.keys()
                    and not self.experiment.metrics[
                        metric_name
                    ].is_reconverable_fetch_e(metric_fetch_e=metric_fetch_e)
                ):
                    status = self._mark_err_trial_status(
                        trial=self.experiment.trials[trial_index],
                        metric_name=metric_name,
                        metric_fetch_e=metric_fetch_e,
                    )
                    self.logger.warning(
                        f"MetricFetchE INFO: Because {metric_name} is an objective, "
                        f"marking trial {trial_index} as {status}."
                    )
                    self._num_trials_bad_due_to_err += 1
                    continue

                self.logger.info(
                    "MetricFetchE INFO: Continuing optimization even though "
                    "MetricFetchE encountered."
                )
                continue

        return results

    def _report_metric_fetch_e(
        self,
        trial: BaseTrial,
        metric_name: str,
        metric_fetch_e: MetricFetchE,
    ) -> None:
        pass

    def _mark_err_trial_status(
        self,
        trial: BaseTrial,
        metric_name: str | None = None,
        metric_fetch_e: MetricFetchE | None = None,
    ) -> TrialStatus:
        trial.mark_failed(unsafe=True)

        return TrialStatus.FAILED

    def _get_failure_rate_exceeded_error(
        self,
        num_bad_in_orchestrator: int,
        num_ran_in_orchestrator: int,
    ) -> FailureRateExceededError:
        return FailureRateExceededError(
            FAILURE_EXCEEDED_MSG.format(
                f_rate=self.options.tolerated_trial_failure_rate,
                n_failed=num_bad_in_orchestrator,
                n_ran=num_ran_in_orchestrator,
                min_failed=self.options.min_failed_trials_for_failure_rate_check,
            )
        )

    def _warn_if_non_terminal_trials(self) -> None:
        """Warns if there are any non-terminal trials on the experiment."""
        non_terminal_trials = [
            t.index for t in self.experiment.trials.values() if not t.status.is_terminal
        ]
        if len(non_terminal_trials) > 0:
            self.logger.warning(
                f"Found {len(non_terminal_trials)} non-terminal trials on "
                f"{self.experiment.name}: {non_terminal_trials}."
            )

    def _should_stop_due_to_global_stopping_strategy(self) -> tuple[bool, str]:
        """Check if optimization should stop due to global stopping strategy."""
        if (
            self.__ignore_global_stopping_strategy
            or self.options.global_stopping_strategy is None
        ):
            return False, ""
        gss = none_throws(self.options.global_stopping_strategy)
        num_trials = len(self.trials)
        if num_trials > 1000:
            self.logger.info(
                f"There are {num_trials} trials; performing "
                f"completion criterion check with {gss}..."
            )
        stop_optimization, global_stopping_msg = gss.should_stop_optimization(
            experiment=self.experiment
        )
        return stop_optimization, global_stopping_msg

    def _should_stop_due_to_total_trials(self) -> tuple[bool, str]:
        """Check if optimization should stop due to total number of trials."""
        num_trials = len(self.trials)
        should_stop = num_trials >= none_throws(self.options.total_trials)
        return (
            should_stop,
            "Exceeding the total number of trials." if should_stop else "",
        )


def get_fitted_adapter(
    orchestrator: Orchestrator, force_refit: bool = False
) -> Adapter:
    """Returns a fitted Adapter object. If the model is fit already, directly
    returns the already fitted model. Otherwise, fits and returns a new one.

    Args:
        orchestrator:The Orchestrator object from which to get the fitted model.
        force_refit: If True, will force a data lookup and a refit of the model.

    Returns:
        An ``Adapter`` object fitted to the observations of the Orchestrator's
        experiment.
    """
    gs = orchestrator.generation_strategy
    adapter = gs.adapter  # Optional[Adapter]
    if adapter is None or force_refit:  # Need to re-fit the model.
        gs._curr._fit(experiment=orchestrator.experiment)
        adapter = cast(Adapter, gs.adapter)
    return adapter



#!/usr/bin/env fbpython
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


import functools
import time
from logging import WARN
from queue import Queue
from threading import Event, Lock

import numpy as np
from ax.adapter.registry import Generators

from ax.core.types import TEvaluationOutcome, TParameterization
from ax.generation_strategy.generation_strategy import (
    GenerationStep,
    GenerationStrategy,
)
from ax.service.ax_client import AxClient
from ax.service.interactive_loop import (
    ax_client_data_attacher,
    interactive_optimize,
    interactive_optimize_with_client,
)
from ax.service.utils.instantiation import ObjectiveProperties
from ax.utils.common.testutils import TestCase
from ax.utils.measurement.synthetic_functions import hartmann6


class TestInteractiveLoop(TestCase):
    def setUp(self) -> None:
        generation_strategy = GenerationStrategy(
            steps=[
                GenerationStep(
                    generator=Generators.SOBOL, max_parallelism=1, num_trials=-1
                )
            ]
        )
        self.ax_client = AxClient(generation_strategy=generation_strategy)
        self.ax_client.create_experiment(
            name="hartmann_test_experiment",
            # pyre-fixme[6]
            parameters=[
                {
                    "name": f"x{i}",
                    "type": "range",
                    "bounds": [0.0, 1.0],
                }
                for i in range(1, 7)
            ],
            objectives={"hartmann6": ObjectiveProperties(minimize=True)},
            tracking_metric_names=["l2norm"],
        )

    def _elicit(
        self,
        parameterization_with_trial_index: tuple[TParameterization, int],
    ) -> tuple[int, TEvaluationOutcome] | None:
        parameterization, trial_index = parameterization_with_trial_index
        x = np.array([parameterization.get(f"x{i + 1}") for i in range(6)])

        return (
            trial_index,
            {
                "hartmann6": (hartmann6(x), 0.0),
                "l2norm": (np.sqrt((x**2).sum()), 0.0),
            },
        )

    def test_interactive_loop(self) -> None:
        optimization_completed = interactive_optimize_with_client(
            ax_client=self.ax_client,
            num_trials=15,
            candidate_queue_maxsize=3,
            # pyre-fixme[6]
            elicitation_function=self._elicit,
        )

        self.assertTrue(optimization_completed)
        self.assertEqual(len(self.ax_client.experiment.trials), 15)

    def test_interactive_loop_aborted(self) -> None:
        # Abort from elicitation function
        def _aborted_elicit(
            parameterization_with_trial_index: tuple[TParameterization, int],
        ) -> tuple[int, TEvaluationOutcome] | None:
            return None

        optimization_completed = interactive_optimize_with_client(
            ax_client=self.ax_client,
            num_trials=15,
            candidate_queue_maxsize=3,
            # pyre-fixme[6]
            elicitation_function=_aborted_elicit,
        )
        self.assertFalse(optimization_completed)

        # Abort from candidate_generator
        def ax_client_candidate_generator(
            queue: Queue[tuple[TParameterization, int] | None],
            stop_event: Event,
            num_trials: int,
            lock: Lock,
        ) -> None:
            with lock:
                queue.put(None)
                stop_event.set()

        ax_client_lock = Lock()
        optimization_completed = interactive_optimize(
            num_trials=15,
            candidate_queue_maxsize=3,
            candidate_generator_function=ax_client_candidate_generator,
            candidate_generator_kwargs={"lock": ax_client_lock},
            data_attacher_function=ax_client_data_attacher,
            data_attacher_kwargs={"ax_client": self.ax_client, "lock": ax_client_lock},
            elicitation_function=self._elicit,
        )
        self.assertFalse(optimization_completed)

    def test_candidate_pregeneration_errors_raised(self) -> None:
        def _sleep_elicit(
            parameterization_with_trial_index: tuple[TParameterization, int],
        ) -> tuple[int, TEvaluationOutcome]:
            parameterization, trial_index = parameterization_with_trial_index
            time.sleep(0.15)  # Sleep to induce MaxParallelismException in loop

            x = np.array([parameterization.get(f"x{i + 1}") for i in range(6)])

            return (
                trial_index,
                {
                    "hartmann6": (hartmann6(x), 0.0),
                    "l2norm": (np.sqrt((x**2).sum()), 0.0),
                },
            )

        with self.assertLogs(logger="ax", level=WARN) as logger:
            interactive_optimize_with_client(
                ax_client=self.ax_client,
                num_trials=3,
                candidate_queue_maxsize=3,
                # pyre-fixme[6]
                elicitation_function=_sleep_elicit,
            )

            # Assert sleep and retry warning is somewhere in the logs
            self.assertIn(
                "sleeping for 0.1 seconds and trying again.",
                functools.reduce(lambda left, right: left + right, logger.output),
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Mapping
from dataclasses import dataclass

import torch
from ax.benchmark.benchmark_problem import BenchmarkProblem, get_soo_opt_config
from ax.benchmark.benchmark_test_function import BenchmarkTestFunction
from ax.core.parameter import ChoiceParameter, ParameterType, RangeParameter
from ax.core.search_space import HierarchicalSearchSpace
from pyre_extensions import none_throws

JENATTON_OPTIMAL_VALUE = 0.1
# Baseline value was obtained with `compute_baseline_value_from_sobol`
JENATTON_BASELINE_VALUE = 0.5797074938368603


def jenatton_test_function(
    x1: int | None = None,
    x2: int | None = None,
    x3: int | None = None,
    x4: float | None = None,
    x5: float | None = None,
    x6: float | None = None,
    x7: float | None = None,
    r8: float | None = None,
    r9: float | None = None,
) -> float:
    """Jenatton test function for hierarchical search spaces.

    This function is taken from:

    R. Jenatton, C. Archambeau, J. González, and M. Seeger. Bayesian
    optimization with tree-structured dependencies. ICML 2017.
    """
    if x1 == 0:
        if x2 == 0:
            return none_throws(x4) ** 2 + 0.1 + none_throws(r8)
        return none_throws(x5) ** 2 + 0.2 + none_throws(r8)
    if x3 == 0:
        return none_throws(x6) ** 2 + 0.3 + none_throws(r9)
    return none_throws(x7) ** 2 + 0.4 + none_throws(r9)


@dataclass(kw_only=True)
class Jenatton(BenchmarkTestFunction):
    """Jenatton test function for hierarchical search spaces."""

    # pyre-fixme[14]: Inconsistent override
    def evaluate_true(self, params: Mapping[str, float | int | None]) -> torch.Tensor:
        # pyre-fixme: Incompatible parameter type [6]: In call
        # `jenatton_test_function`, for 1st positional argument, expected
        # `Optional[float]` but got `Union[None, bool, float, int, str]`.
        value = jenatton_test_function(**params)
        return torch.tensor(value, dtype=torch.double)


def get_jenatton_search_space() -> HierarchicalSearchSpace:
    search_space = HierarchicalSearchSpace(
        parameters=[
            ChoiceParameter(
                name="x1",
                parameter_type=ParameterType.INT,
                values=[0, 1],
                dependents={0: ["x2", "r8"], 1: ["x3", "r9"]},
            ),
            ChoiceParameter(
                name="x2",
                parameter_type=ParameterType.INT,
                values=[0, 1],
                dependents={0: ["x4"], 1: ["x5"]},
            ),
            ChoiceParameter(
                name="x3",
                parameter_type=ParameterType.INT,
                values=[0, 1],
                dependents={0: ["x6"], 1: ["x7"]},
            ),
            *[
                RangeParameter(
                    name=f"x{i}",
                    parameter_type=ParameterType.FLOAT,
                    lower=0.0,
                    upper=1.0,
                )
                for i in range(4, 8)
            ],
            RangeParameter(
                name="r8", parameter_type=ParameterType.FLOAT, lower=0.0, upper=1.0
            ),
            RangeParameter(
                name="r9", parameter_type=ParameterType.FLOAT, lower=0.0, upper=1.0
            ),
        ]
    )
    return search_space


def get_jenatton_benchmark_problem(
    num_trials: int = 50,
    observe_noise_sd: bool = False,
    noise_std: float = 0.0,
) -> BenchmarkProblem:
    name = "Jenatton" + ("_observed_noise" if observe_noise_sd else "")
    optimization_config = get_soo_opt_config(
        outcome_names=[name], observe_noise_sd=observe_noise_sd
    )

    return BenchmarkProblem(
        name=name,
        search_space=get_jenatton_search_space(),
        optimization_config=optimization_config,
        test_function=Jenatton(outcome_names=[name]),
        noise_std=noise_std,
        num_trials=num_trials,
        optimal_value=JENATTON_OPTIMAL_VALUE,
        baseline_value=JENATTON_BASELINE_VALUE,
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import warnings

from bisect import bisect_right
from collections.abc import Iterable, Mapping, Sequence
from copy import deepcopy
from logging import Logger
from math import nan
from typing import Any

import numpy as np
import numpy.typing as npt
import pandas as pd
from ax.core.data import _filter_df, Data
from ax.core.types import TMapTrialEvaluation, TTrialEvaluation
from ax.exceptions.core import UnsupportedError
from ax.utils.common.docutils import copy_doc
from ax.utils.common.equality import dataframe_equals
from ax.utils.common.logger import get_logger
from ax.utils.common.serialization import (
    serialize_init_args,
    TClassDecoderRegistry,
    TDecoderRegistry,
)
from pyre_extensions import assert_is_instance

logger: Logger = get_logger(__name__)
MAP_KEY = "step"


class MapData(Data):
    """Class storing mapping-like results for an experiment.

    Data is stored in a dataframe, and auxiliary information is stored in
    DataFrame with column names given by the keys in the passed ``MapKeyInfo``
    objects.

    Mapping-like results occur whenever a metric is reported as a collection
    of results, each element corresponding to a tuple of values.

    The simplest case is a sequence. For instance a time series is
    a mapping from the 1-tuple `(timestamp)` to (mean, sem) results.

    Another example: MultiFidelity results. This is a mapping from
    `(fidelity_feature_1, ..., fidelity_feature_n)` to (mean, sem) results.

    The dataframe is retrieved via the `map_df` property. The data can be stored
    to an external store for future use by attaching it to an experiment using
    `experiment.attach_data()` (this requires a description to be set.)
    """

    DEDUPLICATE_BY_COLUMNS = ["trial_index", "arm_name", "metric_name"]

    _map_df: pd.DataFrame
    _memo_df: pd.DataFrame | None

    def __init__(
        self,
        df: pd.DataFrame | None = None,
        _skip_ordering_and_validation: bool = False,
    ) -> None:
        """Initialize a ``MapData`` object from the given DataFrame and MapKeyInfos.

        Note: ``MapData`` may also be initialized more simply using
        ``MapData.from_df``, which allows for simpler semantics but may be
        unstable.

        Args:
            df: DataFrame with underlying data, and required columns.
            _skip_ordering_and_validation: If True, uses the given DataFrame
                as is, without ordering its columns or validating its contents.
                Intended only for use in `MapData.filter`, where the contents
                of the DataFrame are known to be ordered and valid.
        """
        map_key_to_type = {MAP_KEY: float}

        if df is None:  # If df is None create an empty dataframe with appropriate cols
            columns = list(self.required_columns().union({MAP_KEY}))
            # Create columns with expected dtypes
            dtype_dict = {**self.COLUMN_DATA_TYPES, **map_key_to_type}

            self._map_df = pd.DataFrame.from_dict(
                {col: pd.Series([], dtype=dtype_dict[col]) for col in columns}
            )
        elif _skip_ordering_and_validation:
            self._map_df = df
        else:
            if MAP_KEY not in df.columns:
                df[MAP_KEY] = nan
            columns = set(df.columns)
            missing_columns = self.required_columns() - columns
            if missing_columns:
                raise ValueError(
                    f"Dataframe must contain required columns {missing_columns}."
                )
            supported_columns = self.supported_columns(extra_column_names=[MAP_KEY])
            extra_columns = columns - supported_columns
            if extra_columns:
                raise UnsupportedError(f"Columns {extra_columns} are not supported.")

            if df["trial_index"].isnull().any():
                df = df.dropna(axis=0, how="all", ignore_index=True)
            else:
                # Don't do this in place so that we now have a copy, so we won't
                # mutate the original df
                df = df.reset_index(drop=True)

            self._map_df = self._safecast_df(df=df, extra_column_types=map_key_to_type)

            col_order = [
                c
                for c in self.column_data_types(extra_column_types=map_key_to_type)
                if c in df.columns
            ]
            if not (self._map_df.columns == col_order).all():
                self._map_df = self._map_df.reindex(columns=col_order)

        self._memo_df = None

    def __eq__(self, o: MapData) -> bool:
        return dataframe_equals(self.map_df, o.map_df)

    @property
    def true_df(self) -> pd.DataFrame:
        return self.map_df

    def required_columns(self) -> set[str]:
        return super().required_columns().union({MAP_KEY})

    @staticmethod
    def from_multiple_map_data(data: Sequence[MapData]) -> MapData:
        if len(data) == 0:
            return MapData()

        # Avoid concatenating empty dataframes which logs a warning.
        non_empty_dfs = [datum.map_df for datum in data if not datum.map_df.empty]
        df = (
            pd.concat(non_empty_dfs)
            if len(non_empty_dfs) > 0
            else pd.DataFrame(
                columns=[*{col for datum in data for col in datum.required_columns()}]
            )
        )

        return MapData(df=df)

    @staticmethod
    def from_map_evaluations(
        evaluations: Mapping[str, TMapTrialEvaluation], trial_index: int
    ) -> MapData:
        records = [
            {
                "arm_name": name,
                "metric_name": metric_name,
                "mean": value[0] if isinstance(value, tuple) else value,
                "sem": value[1] if isinstance(value, tuple) else None,
                "trial_index": trial_index,
                MAP_KEY: step,
            }
            for name, map_dict_and_metrics_list in evaluations.items()
            for step, evaluation in map_dict_and_metrics_list
            for metric_name, value in evaluation.items()
        ]
        return MapData(df=pd.DataFrame(records))

    @property
    def map_df(self) -> pd.DataFrame:
        return self._map_df

    @map_df.setter
    # pyre-fixme[3]: Return type must be annotated.
    def map_df(self, df: pd.DataFrame):
        raise UnsupportedError(
            "MapData's underlying DataFrame is immutable; create a new"
            + " MapData via `__init__` or `from_multiple_data`."
        )

    @classmethod
    def from_multiple_data(cls, data: Iterable[Data]) -> MapData:
        """Downcast instances of Data into instances of MapData with empty
        map_key_infos if necessary then combine as usual (filling in empty cells with
        default values).
        """
        map_datas = [
            (cls(df=datum.df) if not isinstance(datum, MapData) else datum)
            for datum in data
        ]

        return cls.from_multiple_map_data(data=map_datas)

    @property
    def df(self) -> pd.DataFrame:
        """Returns a DataFrame that only contains the last (determined by map keys)
        observation for each (arm, metric) pair.
        """
        if self._memo_df is not None:
            return self._memo_df

        self._memo_df = _tail(map_df=self.map_df, n=1, sort=True)
        return self._memo_df

    @copy_doc(Data.filter)
    def filter(
        self,
        trial_indices: Iterable[int] | None = None,
        metric_names: Iterable[str] | None = None,
    ) -> MapData:
        return MapData(
            df=_filter_df(
                df=self.map_df, trial_indices=trial_indices, metric_names=metric_names
            ),
            _skip_ordering_and_validation=True,
        )

    @classmethod
    def serialize_init_args(cls, obj: Any) -> dict[str, Any]:
        map_data = assert_is_instance(obj, MapData)
        properties = serialize_init_args(
            obj=map_data, exclude_fields=["_skip_ordering_and_validation"]
        )
        properties["df"] = map_data.map_df
        return properties

    @classmethod
    def deserialize_init_args(
        cls,
        args: dict[str, Any],
        decoder_registry: TDecoderRegistry | None = None,
        class_decoder_registry: TClassDecoderRegistry | None = None,
    ) -> dict[str, Any]:
        """Given a dictionary, extract the properties needed to initialize the metric.
        Used for storage.

        Most logic here is for backwards compatibility with older MapData that
        may have been stored with multiple map keys and/or a map key with a
        different name.
        """
        # map_key_infos used to be a supported argument; it allowed the column
        # called MAP_KEY to have a different name.
        if "map_key_infos" in args:
            map_keys = {d["key"] for d in args["map_key_infos"]}
        else:
            map_keys = set()

        deserialized = super().deserialize_init_args(args=args)

        bad_keys = map_keys - {MAP_KEY}
        if len(bad_keys) > 0:
            df = deserialized["df"]
            if MAP_KEY in map_keys:
                warnings.warn(
                    f"Received multiple map keys. All except {MAP_KEY}"
                    " will be ignored.",
                    stacklevel=2,
                )
                if df is not None:
                    df.drop(columns=bad_keys, inplace=True)

            else:
                key_to_rename = bad_keys.pop()
                if len(bad_keys) > 0:
                    warnings.warn(
                        "Received multiple map keys. All except for "
                        f"{key_to_rename} will be ignored.",
                        stacklevel=2,
                    )
                    if df is not None:
                        df.drop(columns=bad_keys, inplace=True)

                warnings.warn(
                    f"{key_to_rename} will be renamed to {MAP_KEY} on "
                    "df, since passing custom map keys is no longer supported.",
                    stacklevel=2,
                )
                if df is not None:
                    df.rename(columns={key_to_rename: MAP_KEY}, inplace=True)

        return deserialized

    def clone(self) -> MapData:
        """Returns a new ``MapData`` object with the same underlying dataframe
        and map key infos.
        """
        return MapData(df=deepcopy(self.map_df))

    def latest(self, rows_per_group: int = 1) -> MapData:
        """Return a new MapData with the most recently observed `rows_per_group`
        rows for each (arm, metric) group, determined by the "step" values,
        where higher implies more recent.

        This function considers only the relative ordering of the "step" values,
        making it most suitable when these values are equally spaced.

        If `rows_per_group` is greater than the number of rows in a given
        (arm, metric) group, then all rows are returned.
        """
        return MapData(
            df=_tail(map_df=self.map_df, n=rows_per_group, sort=True),
        )

    def subsample(
        self,
        keep_every: int | None = None,
        limit_rows_per_group: int | None = None,
        limit_rows_per_metric: int | None = None,
        include_first_last: bool = True,
    ) -> MapData:
        """Return a new MapData that subsamples the `MAP_KEY` column in an
        equally-spaced manner. This function considers only the relative ordering
        of the `MAP_KEY` values, making it most suitable when these values are
        equally spaced.

        There are three ways that this can be done:
            1. If `keep_every = k` is set, then every kth row of the DataFrame in the
                `MAP_KEY` column is kept after grouping by `DEDUPLICATE_BY_COLUMNS`.
                In other words, every kth step of each (arm, metric) will be kept.
            2. If `limit_rows_per_group = n`, the method will find the (arm, metric)
                pair with the largest number of rows in the `MAP_KEY` column and select
                an appropriate `keep_every` such that each (arm, metric) has at most
                `n` rows in the `MAP_KEY` column.
            3. If `limit_rows_per_metric = n`, the method will select an
                appropriate `keep_every` such that the total number of rows per
                metric is less than `n`.
        If multiple of `keep_every`, `limit_rows_per_group`, `limit_rows_per_metric`,
        then the priority is in the order above: 1. `keep_every`,
        2. `limit_rows_per_group`, and 3. `limit_rows_per_metric`.

        Note that we want all curves to be subsampled with nearly the same spacing.
        Internally, the method converts `limit_rows_per_group` and
        `limit_rows_per_metric` to a `keep_every` quantity that will satisfy the
        original request.

        When `include_first_last` is True, then the method will use the `keep_every`
        as a guideline and for each group, produce (nearly) evenly spaced points that
        include the first and last points.
        """
        if (
            keep_every is None
            and limit_rows_per_group is None
            and limit_rows_per_metric is None
        ):
            logger.warning(
                "None of `keep_every`, `limit_rows_per_group`, or "
                "`limit_rows_per_metric` is specified. Returning the original data "
                "without subsampling."
            )
            return self
        subsampled_metric_dfs = []
        for metric_name in self.map_df["metric_name"].unique():
            metric_map_df = _filter_df(self.map_df, metric_names=[metric_name])
            subsampled_metric_dfs.append(
                _subsample_one_metric(
                    metric_map_df,
                    keep_every=keep_every,
                    limit_rows_per_group=limit_rows_per_group,
                    limit_rows_per_metric=limit_rows_per_metric,
                    include_first_last=include_first_last,
                )
            )
        subsampled_df: pd.DataFrame = pd.concat(subsampled_metric_dfs)
        return MapData(df=subsampled_df)

    @classmethod
    def from_evaluations(
        cls,
        evaluations: Mapping[str, TTrialEvaluation],
        trial_index: int,
        sample_sizes: Mapping[str, int] | None = None,
        start_time: int | str | None = None,
        end_time: int | str | None = None,
    ) -> MapData:
        """Not supported for MapData."""
        raise UnsupportedError(
            "MapData.from_evaluations is not supported. "
            "Please use MapData.from_map_evaluations instead."
        )


def _ceil_divide(
    a: int | np.int_ | npt.NDArray[np.int_], b: int | np.int_ | npt.NDArray[np.int_]
) -> np.int_ | npt.NDArray[np.int_]:
    return -np.floor_divide(-a, b)


def _subsample_rate(
    map_df: pd.DataFrame,
    keep_every: int | None = None,
    limit_rows_per_group: int | None = None,
    limit_rows_per_metric: int | None = None,
) -> int:
    if keep_every is not None:
        return keep_every

    grouped_map_df = map_df.groupby(MapData.DEDUPLICATE_BY_COLUMNS)
    group_sizes = grouped_map_df.size()
    max_rows = group_sizes.max()

    if limit_rows_per_group is not None:
        return _ceil_divide(max_rows, limit_rows_per_group).item()

    if limit_rows_per_metric is not None:
        # search for the `keep_every` such that when you apply it to each group,
        # the total number of rows is smaller than `limit_rows_per_metric`.
        ks = np.arange(max_rows, 0, -1)
        # total sizes in ascending order
        total_sizes = np.sum(
            _ceil_divide(group_sizes.values, ks[..., np.newaxis]), axis=1
        )
        # binary search
        i = bisect_right(total_sizes, limit_rows_per_metric)
        # if no such `k` is found, then `derived_keep_every` stays as 1.
        if i > 0:
            return ks[i - 1].item()

    raise ValueError(
        "at least one of `keep_every`, `limit_rows_per_group`, "
        "or `limit_rows_per_metric` must be specified."
    )


def _tail(
    map_df: pd.DataFrame,
    n: int = 1,
    sort: bool = True,
) -> pd.DataFrame:
    """
    Note: Normally, a groupby-apply automatically returns a DataFrame that is
    sorted by the group keys, but this is not true when using filtrations like
    "tail."

    Note: Optimizer beware: This is slow and it has proven difficult to speed it
    up. `tail` takes up a large portion of the time, and so does the groupby;
    sorting can take ~40% of the time. If you find this to be a bottleneck, it
    may be better to avoid unnecessary calls to `.df`.
    """
    if len(map_df) == 0:
        return map_df
    df = map_df.sort_values(MAP_KEY).groupby(MapData.DEDUPLICATE_BY_COLUMNS).tail(n)
    if sort:
        df.sort_values(MapData.DEDUPLICATE_BY_COLUMNS, inplace=True)
    return df


def _subsample_one_metric(
    map_df: pd.DataFrame,
    keep_every: int | None = None,
    limit_rows_per_group: int | None = None,
    limit_rows_per_metric: int | None = None,
    include_first_last: bool = True,
) -> pd.DataFrame:
    """Helper function to subsample a dataframe that holds a single metric."""

    grouped_map_df = map_df.groupby(MapData.DEDUPLICATE_BY_COLUMNS)

    derived_keep_every = _subsample_rate(
        map_df, keep_every, limit_rows_per_group, limit_rows_per_metric
    )

    if derived_keep_every <= 1:
        filtered_map_df = map_df
    else:
        filtered_dfs = []
        for _, df_g in grouped_map_df:
            df_g = df_g.sort_values(MAP_KEY)
            if include_first_last:
                rows_per_group = _ceil_divide(len(df_g), derived_keep_every)
                linspace_idcs = np.linspace(0, len(df_g) - 1, rows_per_group)
                idcs = np.round(linspace_idcs).astype(int)
                filtered_df = df_g.iloc[idcs]
            else:
                filtered_df = df_g.iloc[:: int(derived_keep_every)]
            filtered_dfs.append(filtered_df)
        filtered_map_df: pd.DataFrame = pd.concat(filtered_dfs)
    return filtered_map_df
